import asyncio
import json
import os
import re
from pathlib import Path
import fandom
from span_marker import SpanMarkerModel

span_model_name = "tomaarsen/span-marker-bert-base-fewnerd-fine-super"
sm_model = SpanMarkerModel.from_pretrained(span_model_name).cuda()

def clean_text(text: str) -> str:
    """Clean text until the first unwanted section appears, then stop."""
    cleaned_lines = []
    unwanted_sections = {"Footnotes", "References", "Videos", "Gallery", "External links", "See also", "Other"}

    for line in text.split("\n"):
        line = line.strip()

        # hard stop if line starts with any unwanted section
        if any(line.startswith(sec) for sec in unwanted_sections):
            break

        # remove JSON-LD / schema.org lines
        if line.startswith("{\"@context\":") or line.startswith("{ \"@context\":"):
            continue

        # remove footnotes like ↑ 6.0 6.1 Time of Contempt
        if re.match(r"^↑\s?\d+(\.\d+)*.*", line):
            continue

        # remove empty lines
        if not line:
            continue

        cleaned_lines.append(line)

    return "\n".join(cleaned_lines)

def classify_entities(text: str):
    ents = sm_model.predict(text)
    suggestions = set()
    for e in ents:
        if e["score"] < 0.9:
            continue
        suggestions.add(e["span"].strip())
    return suggestions

def extract_instructions(text: str):
    """Extract instructions from a text and remove them from the main text."""
    lines = text.split('\n')
    capture = False
    extracted_lines = []
    remaining_lines = []

    for i, line in enumerate(lines):
        if i > 0 and lines[i-1].startswith('Quick Answers'):
            capture = True
        elif line.startswith('{'):
            capture = False

        if capture:
            extracted_lines.append(line)
        else:
            remaining_lines.append(line)

    # parse questions/answers
    questions, answers = [], []
    capture = False
    for l in extracted_lines:
        if l.startswith('						Provided by:'):
            capture = False
        if capture:
            answers[-1].append(l)
        if l.endswith('?'):
            questions.append(l)
            answers.append([])
            capture = True

    ans_clean = []
    for a in answers:
        a = ''.join(a)
        a = a.replace('\n', ' ').replace('\t', '').replace("'", '')
        ans_clean.append(a)

    q_a_dict = {q: a for q, a in zip(questions, ans_clean)}
    if len(questions) > 0:
        return q_a_dict, '\n'.join(remaining_lines)
    else:
        return None, text


async def fetch_page(title: str, out_path: Path, instruct_path: Path, search_counter: dict, q: asyncio.Queue, queued: set, visited: set):
    """Download all pages returned by search for a title."""
    try:
        results = await asyncio.to_thread(fandom.search, title)
        search_counter["count"] += 1
        search_counter["queue_size"] = q.qsize()
        print(f"[SEARCH #{search_counter['count']}] {title} -> queue size: {search_counter['queue_size']}")

        if not results:
            print(f"[MISS] {title}")
            return None

        for (page_title, _) in results:
            if page_title in visited:
                continue
            try:
                page = await asyncio.to_thread(fandom.page, page_title)
                text = getattr(page, "plain_text", "")
                if not text:
                    continue

                instructions, cleaned_text = extract_instructions(text)
                cleaned_text = clean_text(cleaned_text)

                fname = page_title.replace("/", "_") + ".txt"
                fpath = out_path / fname
                if not fpath.exists():
                    await asyncio.to_thread(fpath.write_text, cleaned_text, encoding="utf-8")

                if instructions:
                    instruct_file = instruct_path / f"{page_title}.json"
                    def write_json():
                        if not instruct_file.exists():
                            with open(instruct_file, "w", encoding="utf-8") as f:
                                json.dump(instructions, f, ensure_ascii=False, indent=2)
                    await asyncio.to_thread(write_json)
                    print(f"[INSTRUCT] {len(instructions)} instructions saved for {page_title}")

                print(f"[FETCH] {page_title}")
                visited.add(page_title)

            except Exception as inner_e:
                print(f"[ERROR] Failed page {page_title}: {inner_e}")

        return True

    except Exception as e:
        print(f"[ERROR] {title}: {e}")
        return None


async def fetch_worker(q: asyncio.Queue, out_path: Path, instruct_path: Path, visited: set, search_counter: dict, queued: set):
    while True:
        title = await q.get()
        queued.discard(title)
        if title in visited:
            q.task_done()
            continue
        res = await fetch_page(title, out_path, instruct_path, search_counter, q, queued, visited)
        if res:
            visited.add(res)
        q.task_done()


async def classify_worker(out_path: Path, visited: set, q: asyncio.Queue, done: set, queued: set):
    while True:
        for fname in os.listdir(out_path):
            if not fname.endswith(".txt") or fname in done:
                continue
            fpath = out_path / fname
            text = fpath.read_text(encoding="utf-8")
            suggestions = classify_entities(text)
            print(f"[NER] {fname}: {len(suggestions)} suggestions")
            for s in suggestions:
                if s not in visited and s not in queued:
                    queued.add(s)
                    await q.put(s)
                    print(f"[QUEUE] Added '{s}' -> queue size: {q.qsize()}")
            done.add(fname)
        await asyncio.sleep(5)


async def _main(in_path: Path, out_path: Path, instruct_path: Path, n_workers: int = 50):

    out_path.mkdir(parents=True, exist_ok=True)
    instruct_path.mkdir(parents=True, exist_ok=True)

    seeds = []
    for file in os.listdir(in_path):
        if not file.endswith(".json"):
            continue
        with open(in_path / file, "r", encoding="utf-8") as f:
            data = json.load(f)
            if isinstance(data, list):
                seeds.extend([str(x) for x in data])
    print(f"[SEEDS] {len(seeds)}")

    q = asyncio.Queue()
    visited, done = set(), set()
    queued = set()

    for s in seeds:
        if s not in visited and s not in queued:
            queued.add(s)
            await q.put(s)

    search_counter = {'count': 0, 'queue_size': 0}
    fetchers = [asyncio.create_task(fetch_worker(q, out_path, instruct_path, visited, search_counter, queued)) for _ in range(n_workers)]
    classifier = asyncio.create_task(classify_worker(out_path, visited, q, done, queued))

    await asyncio.gather(*fetchers, classifier)


def scrape_fandom(in_path: Path,
                out_path: Path,
                instruct_path: Path,
                n_workers: int = 50,
                wiki: str = "Witcher",
                lang: str = "en"):

    fandom.set_wiki(wiki)
    fandom.set_lang(lang)

    asyncio.run(_main(in_path, out_path, instruct_path, n_workers))



from ..LLM_pieces import DynamicTanh, Dynamic_erf
from dataclasses import dataclass
import torch.nn as nn
from .modules import CirillaBaseModel
from .blocks import Encoder, EncoderArgs, InputEmbeddings
import torch
from einops.layers.torch import Rearrange

@dataclass
class BertArgs(EncoderArgs):
    vocab_size:int = 50_000
    output_what:bool = 'meanpool' # 'meanpool' or 'tokens' or 'vocab' or 'classify'
    cls_index:int = None
    n_classes:int = 2
    tie_params:bool = False
    out_bias:bool = True

    def __post_init__(self):
        assert self.output_what in ['meanpool', 'tokens', 'vocab', 'classify']

class CirillaBERT(
            nn.Module,
            CirillaBaseModel,
            pipeline_tag="text-generation",
            library_name="pytorch",
            license="mit"
    ):
    def __init__(self, args:BertArgs=None):
        super().__init__()

        if isinstance(args, dict):
            args = BertArgs(**args)

        if args is None:
            args = BertArgs()

        self.args = args
        self._prepare_model()

    def _prepare_model(self):

        self.emb = InputEmbeddings(self.args)
        if self.args.layer_norm == "RMSNorm":
            self.layer_norm = nn.RMSNorm(self.args.dim)
        elif self.args.layer_norm == "Derf":
            self.layer_norm = Dynamic_erf(self.args.dim)
        elif self.args.layer_norm == "DyT":
            self.layer_norm = DynamicTanh(self.args.dim)
        else:
            raise ValueError(f"allowed layer norms: 'RMSNorm', 'Derf', 'DyT' ; got: {self.args.layer_norm}")
        self.encoder = Encoder(self.args)

        if self.args.output_what == 'vocab':

            self.output = nn.Linear(self.args.dim, self.args.vocab_size, bias=self.args.out_bias)
            if self.args.tie_params:
                self.output.weight = self.emb.embeddings.weight

        elif self.args.output_what == 'classify':
            if self.args.n_classes == 1:
                self.output = nn.Sequential(nn.Linear(self.args.dim, 1, bias=self.args.out_bias), nn.Sigmoid(), Rearrange('... 1 -> ...'))
            else:
                self.output = nn.Linear(self.args.dim, self.args.n_classes, bias=self.args.out_bias)

        self.n_params = sum(p.numel() for p in self.parameters() if p.requires_grad)

        self.to(self.args.device, dtype=self.args.dtype)
        
    def pred(self, x, attention_mask=None) -> torch.Tensor:
        
        x = self.emb(x)

        if self.args.output_moe_weights:
            x, moe_weights = self.encoder(x)

        else:
            x = self.encoder(x)

        if self.args.output_what == 'meanpool':
            if self.args.output_moe_weights:
                return self.mean_pooling(x, attention_mask), moe_weights
            
            return self.mean_pooling(x, attention_mask)
        
        if self.args.output_what == 'tokens':
            if self.args.output_moe_weights:
                return x, moe_weights
            
            return x
        
        x = self.layer_norm(x)

        if self.args.output_what == 'classify':
            if self.args.cls_index is None:
                x = self.mean_pooling(x, attention_mask)
            else:
                x = x[:, self.args.cls_index]

        x = self.output(x)

        if self.args.output_moe_weights:
            return x, moe_weights
        
        return x
    
    def forward(self, x, attention_mask=None) -> torch.Tensor:
        return self.pred(x, attention_mask)

def bert_training_step(self, data) -> float: # define a custom training step
    torch.compiler.cudagraph_mark_step_begin()

    out = self.model.pred(data[0], data[1]) # tokens, mask
    loss = self.criterion(out, data[2])
    loss_item = loss.item()
    loss.backward()
    return loss_item

@torch.inference_mode()
def bert_inference_step(self, data) -> float: # define a custom inference step
    out = self.model.pred(data[0], data[1]) # tokens, mask
    loss = self.criterion(out, data[2])
    loss_item = loss.item()
    return loss_item


from cirilla.LLM_pieces import (
    SlidingWindowAttention,
    create_dynamic_block_mask,
    create_static_block_mask,
    sliding_window_causal,
    RoPE,
    )
from cirilla.LLM_pieces.sliding_window_attention import AttentionArgs
import torch
import pytest
from attn_gym.mods import generate_tanh_softcap

@pytest.fixture
def get_rope():
    return RoPE(head_dim=128, seq_len=2048, device='cuda')

def test_static_mask(get_rope):
    SOFT_CAP = 20
    softcap = generate_tanh_softcap(SOFT_CAP, approx=False)
    static_mask = create_static_block_mask(sliding_window_causal, 2048, 2048)
    x = torch.rand((1,2048,128*16), device='cuda', dtype=torch.bfloat16) # (b, seq, head_dim*h)

    attention_layer = SlidingWindowAttention(AttentionArgs(
                                                n_heads=16,
                                                n_kv_heads=4,
                                                dim=128*16,
                                                static_mask=True,
                                                window_size=512,
                                                device='cuda'
                                            ),
                                            rope=get_rope,
                                            mask=static_mask,
                                            score_mod=softcap).to('cuda', dtype=torch.bfloat16)
    # verify forward pass
    out = attention_layer(x)
    print(out.shape) # torch.Size([1, 2048, 2048])
    assert out.shape == (1, 2048, 2048)
    
    # verify backward pass
    loss = out.sum()
    loss.backward()

    grads_ok = all(p.grad is not None for p in attention_layer.parameters() if p.requires_grad)
    if not grads_ok:
        for name, p in attention_layer.named_parameters():
            if p.requires_grad and p.grad is None:
                print(f"Parameter {name} has no gradient.")
        raise AssertionError("Not all parameters have gradients after initialization.")

def test_dynamic_mask(get_rope):
    """" dynamic mask - won't trigger recompilation """
    SOFT_CAP = 20
    softcap = generate_tanh_softcap(SOFT_CAP, approx=False)
    x = torch.rand((1,2048,128*16), device='cuda', dtype=torch.bfloat16) # (b, seq, head_dim*h)
    dynamic_args = AttentionArgs(static_mask=False,
                                n_heads=16,
                                n_kv_heads=4,
                                dim=128*16,
                                window_size=512,
                                device='cuda'
                                )
    attention_layer = SlidingWindowAttention(dynamic_args,
                                            mask=create_dynamic_block_mask,
                                            rope=get_rope,
                                            score_mod=softcap).to('cuda', dtype=torch.bfloat16)
    out = attention_layer(x)
    print(out.shape) # torch.Size([1, 2048, 2048])
    assert out.shape == (1, 2048, 2048)

    loss = out.sum()
    loss.backward()
    grads_ok = all(p.grad is not None for p in attention_layer.parameters() if p.requires_grad)
    if not grads_ok:
        for name, p in attention_layer.named_parameters():
            if p.requires_grad and p.grad is None:
                print(f"Parameter {name} has no gradient.")
        raise AssertionError("Not all parameters have gradients after initialization.")

    for tensor_shape in [(1,512,128*16), (1,256,128*16), (1,2048,128*16)]:

        x = torch.rand(tensor_shape, device='cuda', dtype=torch.bfloat16) # (b, seq, head_dim*h)

        out = attention_layer(x)
        print(out.shape)
        assert out.shape == tensor_shape

        loss = out.sum()
        loss.backward()
        grads_ok = all(p.grad is not None for p in attention_layer.parameters() if p.requires_grad)
        if not grads_ok:
            for name, p in attention_layer.named_parameters():
                if p.requires_grad and p.grad is None:
                    print(f"Parameter {name} has no gradient.")
            raise AssertionError("Not all parameters have gradients after initialization.")

    print(create_dynamic_block_mask.cache_info()) # how many times the mask template was reused
    # CacheInfo(hits=1, misses=3, maxsize=32, currsize=3)
    assert create_dynamic_block_mask.cache_info().hits == 1
    assert create_dynamic_block_mask.cache_info().misses == 3
    assert create_dynamic_block_mask.cache_info().currsize == 3


from .bert_model import CirillaBERT, BertArgs, bert_inference_step, bert_training_step
from .dataloader import JSONLDataset, GenericDataset
from .model import Cirilla, Args
from .modules import benchmark_model_part, load_balancing_loss, CirillaBaseModel, get_optims
from .tokenizer_modules import CirillaTokenizer
from .training import TrainingArgs, CirillaTrainer
from .blocks import (
                    Encoder,
                    EncoderArgs,
                    Decoder,
                    DecoderArgs,
                    MLPMixer1D,
                    MixerArgs,
                    VisionEmbeddingModel,
                    KeylessAttention,
                    InputEmbeddings
                    )
from .trm import CirillaTRM, TRMArgs, trm_training_step, trm_inference_step
from .mtp import CirillaMTP, MTPArgs, mtp_training_step, mtp_inference_step

__all__ = [
            'CirillaBERT',
            'BertArgs',
            'Cirilla',
            'Args',
            'JSONLDataset',
            'GenericDataset',
            'CirillaTokenizer',
            'TrainingArgs',
            'CirillaTrainer',
            'benchmark_model_part',
            'load_balancing_loss',
            'CirillaBaseModel',
            'Encoder',
            'EncoderArgs',
            'Decoder',
            'DecoderArgs',
            'InputEmbeddings',
            'VisionEmbeddingModel',
            'KeylessAttention',
            'CirillaTRM',
            'TRMArgs',
            'MLPMixer1D',
            'MixerArgs'
            'CirillaMTP',
            'MTPArgs'
            'trm_training_step',
            'bert_training_step',
            'bert_inference_step',
            'trm_inference_step',
            'mtp_training_step',
            'mtp_inference_step',
            'get_optims'
        ]

try:
    from .mamba_blocks import (
                        HybridDecoder,
                        HybridDecoderArgs
                        )
    
    from .hybrid_model import HybridCirilla, HybridArgs

    __all__ += [
        "HybridDecoder",
        "HybridDecoderArgs",
        "HybridCirilla",
        "HybridArgs"
    ]
except ImportError:
    pass


from tokenizers import SentencePieceBPETokenizer
from transformers import PreTrainedTokenizerFast, AutoTokenizer
from typing import Iterator, Union
import os
from pathlib import Path
from typing import Iterator

SPECIAL_TOKENS = {'unk_token':'<unk>', 'pad_token':'<pad>', 'mask_token':'<mask>',
                'bos_token':'<sos>', 'eos_token':'<eos>', 'system_token':'<|system|>',
                'assistant_token':'<|assistant|>', 'user_token':'<|user|>', 'class_token':'<cls>'}

class CirillaTokenizer:
    def __init__(self, path:Path=None, hub_url=None):
        self.path = path
        self.hub_url = hub_url

        if path is not None:
            if os.path.exists(path):
                self.tokenizer = self._turn_to_fast(path)
        
        elif hub_url is not None:
            self.tokenizer = AutoTokenizer.from_pretrained(hub_url)

    def train(self, dataset: Union[Iterator[str], Iterator[Iterator[str]]], special_tokens: dict[str, str]=SPECIAL_TOKENS, save_to_path:Path='./tokenizer.json', **kwargs) -> PreTrainedTokenizerFast:
        spm = SentencePieceBPETokenizer()
        spm.train_from_iterator(dataset, special_tokens=list(special_tokens.values()), **kwargs)
        spm.save(str(save_to_path))
        self.tokenizer = self._turn_to_fast(save_to_path, special_tokens=special_tokens)
        return self.tokenizer

    @staticmethod
    def _turn_to_fast(path: Path, special_tokens: dict[str, str] = SPECIAL_TOKENS) -> PreTrainedTokenizerFast:
        tokenizer = PreTrainedTokenizerFast(tokenizer_file=str(path), **special_tokens)

        tok_to_add = []
        for s in special_tokens.values():
            if tokenizer.convert_tokens_to_ids(s) == tokenizer.unk_token_id:
                tok_to_add.append(s)
        if tok_to_add:
            tokenizer.add_tokens(tok_to_add)
            tokenizer.add_special_tokens({k: v for k, v in special_tokens.items() if v in tok_to_add})


        tokenizer.chat_template = (
"""{% for message in messages %}
{% if message['role'] == 'system' %}{{'<|system|>' + message['content']}}
{% elif message['role'] == 'user' %}{{'<|user|>' + message['content']}}
{% elif message['role'] == 'assistant' %}{{'<|assistant|>' + message['content']}}
{% endif %}
{% endfor %}{{ '<|assistant|>' if add_generation_prompt else '' }}"""
                )

        return tokenizer


    def pull_from_hub(self, hub_url):
        self.tokenizer = AutoTokenizer.from_pretrained(hub_url)

    def push_to_hub(self, hub_url):
        self.tokenizer.push_to_hub(hub_url)
    
    def decode(self, tokens, **kwargs):
        return self.tokenizer.decode(tokens, **kwargs)
    
    def encode(self, text, **kwargs):
        return self.tokenizer.encode(text, **kwargs)
    
    def apply_chat_template(self, texts, **kwargs):
        return self.tokenizer.apply_chat_template(texts, **kwargs)
    
    def __call__(self, text, **kwargs):
        return self.tokenizer(text, **kwargs)
    
    def convert_tokens_to_ids(self, tokens):
        return self.tokenizer.convert_tokens_to_ids(tokens)



from torch.utils.data import Dataset
from ..Cirilla_model.dataloader import GenericDataset
import json
import torch
import torch.nn.functional as F

class ProtonetDataset(Dataset, GenericDataset):
    def __init__(self, n_support=None, n_query=None, tokenizer=None, **kwargs):
        super().__init__(**kwargs)

        self.n_support = n_support
        self.n_query = n_query
        self.tokenizer = tokenizer

        self._get_data()

    def _get_data(self):

        self.texts = []
        self.labels = []

        for p in self.path:
            with open(p, 'r') as f:

                for line in f:
                    line = json.loads(line)
                    assert line['data type'] == 'bert', 'Setfit supports only BERT data'

                    text = line['text']

                    if self.prefix_tokens is not None:
                        text = "".join(self.prefix_tokens) + text

                    if self.suffix_tokens is not None:
                        text += "".join(self.suffix_tokens)

                    self.texts.append(text)
                    self.labels.append(line['label'])

        slabels = set(self.labels)
        assert sorted(list(slabels)) == list(range(len(slabels))), "Labels must be consecutive integers starting from 0"
        self.labels = torch.tensor(self.labels, dtype=torch.int64)
        self.n_classes = len(slabels)
        self.label_to_indices = {}

        for s in slabels:
            indices = torch.where(self.labels == s)[0]
            if len(indices) < 2:
                print(f"Label {s} has only {len(indices)} samples. Needs at least 2. It will be skipped.")
                continue

            self.label_to_indices[s] = indices

        if self.n_support is None or self.n_query is None:
            min_examples = min([len(indices) for indices in self.label_to_indices.values()])
            self.n_support = self.n_support if self.n_support is not None else min_examples // 2
            self.n_query = self.n_query if self.n_query is not None else min_examples - self.n_support

        for label, indices in self.label_to_indices.items(): # sanity check
            if len(indices) < self.n_support + self.n_query:
                raise ValueError(f"Label {label} has only {len(indices)} samples. Needs at least {self.n_support + self.n_query}")
                
    def __len__(self):
        return len(self.texts) * self.n_classes
    
    def __getitem__(self, idx):
        
        support_inputs_id = []
        support_inputs_mask = []
        query_inputs_id = []
        query_inputs_mask = []
        support_labels = []
        query_labels = []

        for klass, indices in self.label_to_indices.items():

            # sample without replacement
            perm = torch.randperm(indices.size(0))
            indices = indices[perm][: self.n_support + self.n_query]

            support_indices = indices[:self.n_support]
            query_indices = indices[self.n_support:self.n_support + self.n_query]

            support_texts = [self.texts[i] for i in support_indices]
            query_texts = [self.texts[i] for i in query_indices]

            for text in support_texts:
                tokens_mask = self.tokenizer(text, return_tensors='pt', padding='max_length', truncation=True, max_length=self.max_len)
                
                support_inputs_id.append(tokens_mask['input_ids'].to(self.device))
                support_inputs_mask.append(tokens_mask['attention_mask'].to(self.device))
                support_labels.append(klass)

            for text in query_texts:
                tokens_mask = self.tokenizer(text, return_tensors='pt', padding='max_length', truncation=True, max_length=self.max_len)
                
                query_inputs_id.append(tokens_mask['input_ids'].to(self.device))
                query_inputs_mask.append(tokens_mask['attention_mask'].to(self.device))
                query_labels.append(klass)

        return {
                'support_inputs_id': torch.concat(support_inputs_id),
                'support_inputs_mask': torch.concat(support_inputs_mask),
                'support_labels': torch.tensor(support_labels, dtype=torch.int64, device=self.device),
                'query_inputs_id': torch.concat(query_inputs_id),
                'query_inputs_mask': torch.concat(query_inputs_mask),
                'query_labels': torch.tensor(query_labels, dtype=torch.int64, device=self.device),
        }

def protonet_training_step(self, data):
    torch.compiler.cudagraph_mark_step_begin()
    
    support_inputs_ids = data['support_inputs_id']
    b, sexamples, seq = support_inputs_ids.shape

    support_inputs_ids = support_inputs_ids.view(b*sexamples, seq)

    support_input_masks = data['support_inputs_mask']
    support_input_masks = support_input_masks.view(b*sexamples, seq)

    support_labels = data['support_labels']

    query_inputs_ids = data['query_inputs_id']
    b, qexamples, seq = query_inputs_ids.shape
    query_inputs_ids = query_inputs_ids.view(b*qexamples, seq)

    query_input_masks = data['query_inputs_mask']
    query_input_masks = query_input_masks.view(b*qexamples, seq)

    query_labels = data['query_labels']

    support_emb = self.model(support_inputs_ids, support_input_masks).view(b, sexamples, -1)
    query_emb = self.model(query_inputs_ids, query_input_masks).view(b, qexamples, -1)

    class_prototypes = {}
    for class_id in torch.unique(support_labels):
        a = []
        for b_ in range(b):
            class_mask = support_labels[b_] == class_id
            class_emb = support_emb[b_][class_mask]
            a.append(class_emb.mean(dim=0).unsqueeze(0))

        class_prototypes[class_id.item()] = torch.concat(a)

    distances = torch.empty(b, qexamples, len(class_prototypes), device=getattr(self.model.args, 'device', 'cuda'), dtype=query_emb.dtype)

    for i, class_prototype in enumerate(class_prototypes.values()):
        distances[:, :, i] = - torch.norm(query_emb - class_prototype.unsqueeze(1), dim=2)

    loss = F.cross_entropy(distances.view(b*qexamples, -1), query_labels.view(b*qexamples))
    loss_item = loss.item()
    loss.backward()
    import time
    return loss_item

@torch.inference_mode()
def protonet_inference_step(self, data):
    support_inputs_ids = data['support_inputs_id']
    b, sexamples, seq = support_inputs_ids.shape

    support_inputs_ids = support_inputs_ids.view(b*sexamples, seq)

    support_input_masks = data['support_inputs_mask']
    support_input_masks = support_input_masks.view(b*sexamples, seq)

    support_labels = data['support_labels']

    query_inputs_ids = data['query_inputs_id']
    b, qexamples, seq = query_inputs_ids.shape
    query_inputs_ids = query_inputs_ids.view(b*qexamples, seq)

    query_input_masks = data['query_inputs_mask']
    query_input_masks = query_input_masks.view(b*qexamples, seq)

    query_labels = data['query_labels']

    support_emb = self.model(support_inputs_ids, support_input_masks).view(b, sexamples, -1)
    query_emb = self.model(query_inputs_ids, query_input_masks).view(b, qexamples, -1)

    class_prototypes = {}
    for class_id in torch.unique(support_labels):
        a = []
        for b_ in range(b):
            class_mask = support_labels[b_] == class_id
            class_emb = support_emb[b_][class_mask]
            a.append(class_emb.mean(dim=0).unsqueeze(0))

        class_prototypes[class_id.item()] = torch.concat(a)

    distances = torch.empty(b, qexamples, len(class_prototypes), device=getattr(self.model.args, 'device', 'cuda'), dtype=query_emb.dtype)

    for i, class_prototype in enumerate(class_prototypes.values()):
        distances[:, :, i] = - torch.norm(query_emb - class_prototype.unsqueeze(1), dim=2)

    loss = F.cross_entropy(distances.view(b*qexamples, -1), query_labels.view(b*qexamples))
    loss_item = loss.item()
    
    return loss_item


import ollama
from pydantic import BaseModel
import time
import os
import json
from tqdm import tqdm
from typing import Optional
import copy
from .multi_turn_vllm import random_user_prompts
import random
from pathlib import Path

"""
Ollama options
    numa: Optional[bool] = None
    num_ctx: Optional[int] = None
    num_batch: Optional[int] = None
    num_gpu: Optional[int] = None
    main_gpu: Optional[int] = None
    low_vram: Optional[bool] = None
    f16_kv: Optional[bool] = None
    logits_all: Optional[bool] = None
    vocab_only: Optional[bool] = None
    use_mmap: Optional[bool] = None
    use_mlock: Optional[bool] = None
    embedding_only: Optional[bool] = None
    num_thread: Optional[int] = None

    # runtime options
    num_keep: Optional[int] = None
    seed: Optional[int] = None
    num_predict: Optional[int] = None
    top_k: Optional[int] = None
    top_p: Optional[float] = None
    tfs_z: Optional[float] = None
    typical_p: Optional[float] = None
    repeat_last_n: Optional[int] = None
    temperature: Optional[float] = None
    repeat_penalty: Optional[float] = None
    presence_penalty: Optional[float] = None
    frequency_penalty: Optional[float] = None
    mirostat: Optional[int] = None
    mirostat_tau: Optional[float] = None
    mirostat_eta: Optional[float] = None
    penalize_newline: Optional[bool] = None
    stop: Optional[Sequence[str]] = None
"""

class OllamaCurate:
    def __init__(self, model, system_prompt, response_template:BaseModel):
        self.system_prompt = {'role': 'system', 'content': system_prompt}
        self.model = model
        self.response_template = response_template
        self.convo = {}
    
    @staticmethod
    def try_repair_json(text: str) -> str:
        text = text.strip()
        try:
            json.loads(text)
            return text
        except:
            pass

        if text.count('{') > text.count('}'):
            text += "}"
        if text.count('[') > text.count(']'):
            text += "]"
        if text.count('"') % 2 != 0:
            text += '"'

        try:
            json.loads(text)
            return text
        except:
            return json.dumps({"summary": text[:1000]})


    def single_pass_summary(
        self,
        paths: list[Path],
        save_to: Path = "./summaries",
        seed: int = 42,
        num_predict: int = 4096,  # allow longer completion
        use_response_template: bool = False,
    ):
        """
        Summarize a list of files using a single pass of the model
        
        Args:
            paths (list[Path]): List of paths of files to summarize.
            save_to (Path): Directory to save summaries. Defaults to './summaries'.
            seed (int): Random seed for reproducibility. Defaults to 42.
            num_predict (int): Number of tokens to predict. Defaults to 4096.
            use_response_template (bool: Use a response template. Defaults to False.
        
        Returns:
            None
        """
        os.makedirs(save_to, exist_ok=True)

        def _call_model(prompt: str):
            if use_response_template and hasattr(self, "response_template") and self.response_template is not None:
                resp = ollama.chat(
                    model=self.model,
                    messages=[self.system_prompt, {"role": "user", "content": prompt}],
                    format=self.response_template.model_json_schema(),
                    options={"num_predict": num_predict, "seed": seed},
                )
                raw = resp.message.content.strip()
                raw = self.try_repair_json(raw)
                resp_valid = self.response_template.model_validate_json(raw)
                return resp_valid.summary
            else:
                resp = ollama.chat(
                    model=self.model,
                    messages=[self.system_prompt, {"role": "user", "content": prompt}],
                    options={"num_predict": num_predict, "seed": seed},
                )
                return resp.message.content.strip()

        bar = tqdm(total=len(paths), desc="Processing files", unit="file")

        for path in paths:
            basename = os.path.basename(path).split(".")[0]
            final_path = os.path.join(save_to, f"{basename}.txt")

            if not os.path.exists(final_path):
                with open(path, "r", encoding="utf-8") as f:
                    full_text = f.read()

                prompt = \
f"""Summarize the TEXT into one detailed, single-paragraph summary.

Rules:
- Include ALL concrete details: names, places, monsters, dates, events, factions, motives, artifacts, numbers, outcomes.
- Stay only within Witcher lore (books, games, monsters, history, characters). Ignore anything about Netflix, actors, or production.
- Do not add or invent facts. Use only what is in the text.
- Be factual, clear, and exhaustive while keeping one coherent paragraph.
- If the text has almost no real facts, instead list up to 12 observable items (comma-separated).

TEXT:
{full_text}

"""
                try:
                    final_summary = _call_model(prompt).strip()
                    final_summary = final_summary.replace("{", "").replace("}", "")
                except Exception as e:
                    print(f"Model error on {path}: {e}")
                    final_summary = full_text[:1000]  # truncated fallback

                with open(final_path, "w", encoding="utf-8") as f:
                    f.write(final_summary)

            bar.update(1)

        os.system(f"ollama stop {self.model}")

    def dynamic_hierarchical_summary(
        self,
        paths: list[Path],
        save_to: Path = './summaries',
        chunk_lines: int = 100, # number of lines per chunk
        seed: int = 42,
        num_predict: int = 2048, # max number of tokens
        max_words_summary: int = 500, # target maximum words per summary block
        use_response_template: bool = False
    ):
        """
        Summarize a list of files hierarchically. Meaning that for a text divided into chunks we subsequently generate a summary for each chunk, and then generate a final summary based on the summaries of the chunks.
        
        Args:
            paths (list[Path]): List of paths of files to summarize.
            save_to (Path): Directory to save summaries. Defaults to './summaries'.
            chunk_lines (int): Number of lines per chunk. Defaults to 100.
            seed (int): Random seed for reproducibility. Defaults to 42.
            num_predict (int): Number of tokens to predict. Defaults to 2048.
            max_words_summary (int): Maximum number of words per summary block and the final summary. Defaults to 500.
            use_response_template (bool): Use a response template. Defaults to False.
        
        Returns:
            None
        """
        os.makedirs(save_to, exist_ok=True)

        def _call_model(prompt: str):
            if use_response_template and hasattr(self, 'response_template') and self.response_template is not None:
                resp = ollama.chat(
                    model=self.model,
                    messages=[self.system_prompt, {'role': 'user', 'content': prompt}],
                    format=self.response_template.model_json_schema(),
                    options={'num_predict': num_predict, 'seed': seed}
                )

                raw = resp.message.content.strip()
                raw = self.try_repair_json(raw)

                resp_valid = self.response_template.model_validate_json(raw)
                data = resp_valid.model_dump()
                return data['summary']

            else:
                resp = ollama.chat(
                    model=self.model,
                    messages=[self.system_prompt, {'role': 'user', 'content': prompt}],
                    options={'num_predict': num_predict, 'seed': seed}
                )
                raw_text = resp.message.content.strip()

                return raw_text


        def summarize_chunk(chunk_text: str, context_summary: Optional[str] = None, target_words: int = max_words_summary):
            if context_summary:
                prompt = f"""
    You are an objective, concise summarizer. Use the EXISTING SUMMARY together with the NEW CHUNK to produce a single, self-contained updated summary.
    Rules:
    - Do NOT invent facts or add information not present in the EXISTING SUMMARY or NEW CHUNK.
    - Prioritize new, important facts introduced by the NEW CHUNK; integrate them into the EXISTING SUMMARY and remove redundancies.
    - If the NEW CHUNK adds no substantive facts, look for any noteworthy detail in the NEW CHUNK and highlight it in one sentence; if there is literally nothing (only noise: whitespace, headers, repeated filler), instead output a compact list (comma-separated, max 8 items) of tokens/phrases or types of content that do appear (e.g. "table of contents, URL, image caption").
    - Be factual, concise and clear. Return a single paragraph (no headings, no metadata).
    - Keep length ≤ {target_words} words and avoid repetition.

    EXISTING SUMMARY:
    {context_summary}

    NEW CHUNK:
    {chunk_text}
    """
            else:
                prompt = f"""
    You are an objective, concise summarizer. Summarize the TEXT into a single, self-contained paragraph.
    Rules:
    - Preserve concrete facts, key points, entities, dates, and actions. Do NOT invent details.
    - If the text contains no substantive facts (only noise or empty lines), find the most salient thing present and describe it in one sentence; if there truly is nothing useful, output a compact list (comma-separated, max 8 items) of observable elements or token types (e.g. "heading, URL, bullet list").
    - Prefer clarity and specificity. Return only the summary (no headings or explanations).
    - Keep length ≤ {target_words} words.
    TEXT:
    {chunk_text}
    """

            return _call_model(prompt).strip()

        def _chunks_from_file(path, lines_per_chunk):
            with open(path, 'r', encoding='utf-8') as f:
                all_lines = f.read().splitlines()

            for i in range(0, len(all_lines), lines_per_chunk):
                chunk_lines_list = all_lines[i:i + lines_per_chunk]
                yield "\n".join(chunk_lines_list)

        bar = tqdm(total=len(paths), desc="Processing files", unit="file")
        for file_idx, path in enumerate(paths):
            basename = os.path.basename(path).split('.')[0]

            if not os.path.exists(f"{save_to}/{basename}.txt"):

                summary_blocks = []
                current_summary = None
                processed_chunks = 0

                for chunk_idx, chunk_text in enumerate(_chunks_from_file(path, chunk_lines)):
                    # skip empty chunk (all whitespace)
                    if not chunk_text.strip():
                        continue

                    # summarize (current_summary + chunk_text) into a new current_summary
                    try:
                        if current_summary is None:
                            new_summary = summarize_chunk(chunk_text, context_summary=None)
                        else:
                            # include the running summary as context
                            new_summary = summarize_chunk(chunk_text, context_summary=current_summary)
                    except Exception as e:
                        print(f"Model error on chunk {processed_chunks} ({path} chunk {chunk_idx}): {e}")
                        # fallback: just store the chunk itself as a "summary" to avoid data loss
                        new_summary = chunk_text[:1000]  # truncated fallback

                    block_id = f"{file_idx:03d}_{chunk_idx:03d}"
                    summary_blocks.append({'id': block_id, 'summary': new_summary, 'source': f"{basename}:{chunk_idx}"})
                    current_summary = new_summary  # update running summary
                    processed_chunks += 1

                    # elapsed = time.time() - start
                    # print(f"Processed chunk #{processed_chunks} (block {block_id}) — elapsed {elapsed/60:.2f} min", end='\r')
                if len(summary_blocks) > 1:
                    cohesive_blocks = [b['summary'] for b in summary_blocks]
                    cohesive_blocks = "\n\n".join(cohesive_blocks)

                    prompt = f"""
    You are an objective, concise summarizer. You are given BLOCK SUMMARIES that each describe parts of a larger document. Produce ONE cohesive, comprehensive summary of the entire document.
    Rules:
    - Integrate facts across blocks, remove redundancy, resolve obvious repetition, and create a smooth, logical single-paragraph narrative.
    - Do NOT hallucinate or add information not present in the blocks. If a fact appears in multiple blocks, mention it once.
    - If many blocks are sparse, prioritize blocks that contain specific facts or named entities; if none contain substantive facts, list (comma-separated) the observable items across blocks (max 12).
    - Return only the final summary (no headings). Aim for ≈ {max_words_summary} words; do not exceed that by more than ~20%.
    BLOCK SUMMARIES:
    {cohesive_blocks}
    """

                    final_summary = _call_model(prompt).strip()
                else:
                    final_summary = summary_blocks[0]['summary']

                final_path = os.path.join(save_to, f"{basename}.txt")
                with open(final_path, 'w', encoding='utf-8') as f:
                    f.write(final_summary)

            bar.update(1)

        os.system(f'ollama stop {self.model}')

    def __call__(self, paths:list[Path], save_to:Path='./example', seed:int=42, checkpoint:int=10, skip:bool=True) -> None:
        """
        Generate instructions from files.

        Args:
            paths (list[Path]): List of paths to files to summarize.
            save_to (Path): Directory to save summaries. Defaults to './example'.
            seed (int): Random seed for reproducibility. Defaults to 42.
            checkpoint (int): Number of files to process before saving a checkpoint. Defaults to 10.
            skip (bool): Skip files that already have summaries. Defaults to True. Else, add index to file name.

        Returns:
            None
        """

        start = time.time()
        n_skipped = 0
        n_failed = 0
        not_failed = 0
        
        os.makedirs(save_to, exist_ok=True)

        for i,p in enumerate(paths):

            data = open(p, 'r').read()
            data = {'role': 'user', 'content': data}
            
            chat = [self.system_prompt]

            if '/' in p:
                p = p.split('/')[-1]
            p = p.split('.')[0]

            if os.path.exists(f'{save_to}/{p}.json'):
                if skip:
                    n_skipped += 1
                    continue
                else:
                    with open(f'{save_to}/{p}.json', 'r') as f:
                        qa = json.load(f)
                    
                    questions=f"find a question and answer pair that is different from:\nquestion: {qa['question']} answer: {qa['answer']}\n"
                    p = f"{p}_"
                    if os.path.exists(f'{save_to}/{p}.json'):
                        j=1
                        while True:
                            p_ = f"{p}{j}"
                            if not os.path.exists(f'{save_to}/{p_}.json'):
                                p = p_
                                break
                            j += 1
                            with open(f'{save_to}/{p_}.json', 'r') as f:
                                qa = json.load(f)
                            questions = questions + f"question: {qa['question']} answer: {qa['answer']}\n"

                    chat.append({'role': 'user', 'content': questions})

            chat.append(data)

            response = ollama.chat(
                model=self.model,
                messages = chat,
                format=self.response_template.model_json_schema(),
                options={
                    'num_predict': 512, # max num tokens
                    'seed': seed
                }
            )
            try:
                response = self.response_template.model_validate_json(response.message.content)
                response = response.model_dump()
                self.convo[p] = response
                not_failed += 1
            except:
                n_failed += 1
                print(f'failed on {p} failed:not failed {n_failed}:{not_failed}')
            
            print(f' ETA: {((time.time() - start) / (i+1 - n_skipped) * (len(paths) - i))/60:3.1f} min ', end='\r')

            if (i % checkpoint == 0) or (i == len(paths)-1):
                for k, v in self.convo.items():
                    # if not os.path.exists(f'{save_to}/{k}.json'):
                    with open(f'{save_to}/{k}.json', 'w') as f:
                        json.dump(v, f, indent=2)
                
                self.convo = {}

        os.system(f'ollama stop {self.model}')

    def multi_turn(self,
        paths: list[Path],
        save_to: Path='./convos',
        bar: tqdm=None,
        n_turns_range: tuple[int,int]=(2,5),
        seed: int=random.randint(0, 1000),
        prob_chance_new_context: float=0.3):
        """
        Generate multi turn question answer pairs from files.

        Args:
            paths (list[Path]): List of paths to files to summarize.
            save_to (Path): Directory to save summaries. Defaults to './example'.
            bar (tqdm.tqdm): tqdm bar to track progress. Defaults to None.
            n_turns_range (tuple[int,int]): Range of number of turns to generate. Defaults to (2, 5).
            seed (int): Random seed for reproducibility. Defaults to 42.
            prob_chance_new_context (float): Probability of starting a new context for the question answer pairs. Defaults to 0.3.

        Returns:
            None
        """
    
        sys_prompt = \
    """You are an AI assistant engaged in a multi-turn question answering conversation.  

    Your task:
    - Read and understand the provided content.
    - Generate a clear and relevant **question** about the content.
    - Provide a factual and concise **answer** to that question.  

    Content to analyze:
    {content}

    Conversation rules:
    - You may sometimes ask a very general question about "The Witcher" universe (books, lore, characters, history, monsters, or games), but you must **never** ask or answer about the Netflix adaptation, series, actors, or production.
    - If new content is introduced, you must adapt your question and answer to that **new context**, ensuring it feels like a natural continuation of the conversation.
    - Each response should stay grounded in either the current context or the newly provided content.
    - Do not repeat earlier questions unless the new context requires it.
    - Keep answers informative, but concise and accurate.  
    - You may ask for clarification or a follow-up question. DO NOT ASK THE SAME QUESTION, you may ask to clarify or rephrase instead.
    - If the context is very vague or confusing you can ask about some obvious fact or element present in the context.

    Format:
    Return a JSON object that matches the provided schema with the following keys:
    - "question": the generated question
    - "answer": the corresponding answer

    Already asked questions: (DO NOT REPEAT THEM AGAIN)
    """

        new_content_prompt = \
    """Now, based on the NEW CONTENT below, generate a new question and answer.  

    Rules:
    - The question must be relevant to the NEW CONTENT.
    - The answer must be factual and concise.
    - You may ask a general question about The Witcher universe if it fits naturally, but you must never ask or answer about the Netflix adaptation, series, actors, or production.
    - Do not repeat earlier questions unless the NEW CONTENT explicitly makes it necessary.
    - Ensure the new question and answer feel like a natural continuation of the ongoing conversation.  

    NEW CONTENT:
    {random_content}
    """

        not_failed = 0
        n_failed = 0

        for p in paths:

            basename = os.path.basename(p).split('.')[0]

            with open(p, 'r') as f:
                path_content = f.read()

            clean_convo = []
            model_convo = [{'role': 'system', 'content': sys_prompt.format(content=path_content)}]
            random_path = None

            for i in range(random.randint(n_turns_range[0], n_turns_range[1])):

                if (random.randint(0, 100) < prob_chance_new_context * 100) and (i > 0):
                    random_path = random.choice(paths)
                    with open(random_path, 'r') as f:
                        random_content = f.read()

                    model_convo.append({'role':'user', 'content':new_content_prompt.replace('{random_content}', random_content)})
                    
                else:
                    model_convo.append({'role':'user', 'content':random.choice(random_user_prompts)})
                
                response = ollama.chat(
                    model=self.model,
                    messages = model_convo,
                    format=self.response_template.model_json_schema(),
                    options={
                        'num_predict': 512, # max num tokens
                        'seed': seed
                    }
                )
                try:
                    response = self.response_template.model_validate_json(response.message.content)
                    response = response.model_dump()
                    model_convo.append({'role':'assistant', 'content':f"Question: {response['question']}\nAnswer: {response['answer']}"})
                    response['context'] = basename if random_path is None else os.path.basename(random_path).split('.')[0]
                    if not (response['question'] == "" or response['answer'] == ""):
                        clean_convo.append(response)
                    not_failed += 1
                except:
                    n_failed += 1
                    print(f'failed on {p} failed:not failed {n_failed}:{not_failed}')

            if len(clean_convo) > n_turns_range[0]:

                clean_convo = tuple(clean_convo)

                basename_ = copy.copy(basename)
                i = 0
                while True:
                    if os.path.exists(f'{save_to}/{basename_}.json'):
                        basename_ = copy.copy(basename) + str(i)
                        i += 1
                    else:
                        break
                
                os.makedirs(save_to, exist_ok=True)

                with open(f'{save_to}/{basename_}.json', 'w') as f:
                    json.dump(clean_convo, f)

                if bar is not None:
                    bar.update(1)
            
        os.system(f'ollama stop {self.model}')



from dataclasses import dataclass
import torch.nn as nn
from .modules import select_torch_device, CirillaBaseModel
import torch
from contextlib import nullcontext
from einops.layers.torch import Rearrange
import torch.nn.functional as F
from .blocks import InputEmbeddings

@dataclass
class TRMArgs:
    """general"""
    vocab_size:int = 70
    dim:int = 256
    tie_params:bool = False
    out_bias:bool = True
    
    """misc"""
    dtype_str:str = 'bfloat16'
    n_total_refinements:int = 4
    n_latent_refinements:int = 2
    device:str = select_torch_device()

    @property
    def dtype(self):
        if self.dtype_str == "fp8":
            return torch.bfloat16 # for initialization, then convert to FP8
        return getattr(torch, self.dtype_str)

class CirillaTRM(
            nn.Module,
            CirillaBaseModel,
            pipeline_tag="text-generation",
            library_name="pytorch",
            license="mit"
    ):
    def __init__(self, network:nn.Module, args:TRMArgs=None):
        super().__init__()

        if isinstance(args, dict):
            args = TRMArgs(**args)

        if args is None:
            args = TRMArgs()

        self.args = args

        self.network = network

        self.emb = InputEmbeddings(self.args)

        self.y_hat_init = nn.Parameter(torch.randn(self.args.dim) * 1e-2)
        self.z_init = nn.Parameter(torch.randn(self.args.dim) * 1e-2)

        self.output = nn.Linear(self.args.dim, self.args.vocab_size, bias=self.args.out_bias)
        if self.args.tie_params:
            self.output.weight = self.emb.embeddings.weight

        self.to_halt = nn.Sequential(
                        nn.Linear(self.args.dim, 1),
                        Rearrange('... 1 -> ...')
                        )

        self.n_params = sum(p.numel() for p in self.parameters() if p.requires_grad)

        self.to(self.args.device, dtype=self.args.dtype)

    def get_init(self) -> tuple[torch.Tensor, torch.Tensor]:
        return self.y_hat_init, self.z_init
    
    def get_halt(self, x, attention_mask=None) -> torch.Tensor:
        return self.to_halt(self.mean_pooling(x, attention_mask))
    
    def single_refinement_step(self, x, y_hat, z) -> tuple[torch.Tensor, torch.Tensor]:

        for _ in range(self.args.n_latent_refinements):

            z = self.network(x + y_hat + z)

        y_hat = self.network(y_hat + z)
        return y_hat, z
    
    def refine(self, x, y_hat, z) -> tuple[torch.Tensor, torch.Tensor]:

        for step in range(self.args.n_total_refinements):

            is_last_step = step == self.args.n_total_refinements - 1

            context = torch.no_grad if not is_last_step else nullcontext

            with context():
                y_hat, z = self.single_refinement_step(x, y_hat, z)

        return y_hat, z

    def forward(self, x, y_hat, z, attention_mask=None) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        
        x = self.emb(x)

        y_hat, z = self.refine(x, y_hat, z)

        pred = self.output(y_hat)

        haltp = self.get_halt(y_hat, attention_mask)

        return pred, y_hat, z, haltp
    
    @torch.inference_mode()
    def predict(self, x, attention_mask=None, halt_thresh=0.5, max_recurrent_step=16) -> tuple[torch.Tensor, torch.Tensor]:
        
        y_hat, z = self.get_init()

        preds = []
        pred_indices = []
        n_steps = []
        active_batch_indices = torch.arange(x.shape[0], device=x.device)

        for step in range(max_recurrent_step):

            pred, y_hat, z, haltp = self.forward(x, y_hat, z, attention_mask)

            halt_mask = (F.sigmoid(haltp) < halt_thresh) & (step < max_recurrent_step - 1)

            if halt_mask.all():
                continue
            
            y_hat = y_hat[halt_mask]
            z = z[halt_mask]
            x = x[halt_mask]
            attention_mask = attention_mask[halt_mask]

            preds.append(pred[~halt_mask])
            n_steps.extend([step] * (~halt_mask).sum().item())
            pred_indices.append(active_batch_indices[~halt_mask])
            active_batch_indices = active_batch_indices[halt_mask]

            if z.numel() == 0: # if is empty
                break

        preds = torch.cat(preds, dim=0)
        n_steps = torch.tensor(n_steps).to(x.device)
        pred_indices = torch.cat(pred_indices, dim=0).to(x.device)
        
        preds = preds[pred_indices]
        n_steps = n_steps[pred_indices]

        return preds, n_steps

def trm_training_step(self, data, max_recurrent_step=16, halt_weight=0.5, halt_thresh=0.5, ema_model=None) -> float:

    step_loss = 0
    n = 0

    y_hat, z = self.model.get_init()
    x = data[0]
    mask = data[1]

    for _ in range(max_recurrent_step):

        torch.compiler.cudagraph_mark_step_begin()

        pred, y_hat, z, haltp = self.model(x, y_hat, z, mask)

        loss = F.cross_entropy(pred.view(-1, self.model.args.vocab_size), x.view(-1))

        all_correct = (pred.argmax(dim=-1) == x).all(dim=-1) # here we just want to predict the x itself, this is a trivial task

        halt_loss = F.binary_cross_entropy_with_logits(haltp, all_correct.to(haltp.dtype))

        loss = loss + halt_weight * halt_loss

        step_loss += loss.item()
        n += 1

        loss.backward()
        if ema_model is not None:
            ema_model.update() # model needs to have .predict() method

        halt_mask = F.sigmoid(haltp) < halt_thresh
        
        y_hat, z = y_hat.detach(), z.detach()

        if halt_mask.all():
            continue
        
        y_hat = y_hat[halt_mask]
        z = z[halt_mask]
        x = x[halt_mask]
        mask = mask[halt_mask]

        if z.numel() == 0: # if is empty
            break

    return step_loss / n

@torch.inference_mode()
def trm_inference_step(self, data, max_recurrent_step=16, halt_thresh=0.5) -> float:
    
    x = data[0]
    mask = data[1]

    preds, n_steps = self.model.predict(x, mask, halt_thresh, max_recurrent_step)

    loss = F.cross_entropy(preds.view(-1, self.model.args.vocab_size), x.view(-1))

    return loss.item()


from cirilla.LLM_pieces import RoPE
import pytest
import torch

@pytest.fixture
def get_rope():
    return RoPE(head_dim=128, seq_len=1024, device='cuda:0')

@pytest.mark.parametrize("seq_len", [
    256,
    512,
    1024,
])
def test_rope_initialization(seq_len, get_rope):
    rope = get_rope
    x = torch.randn(2, seq_len, 4, 128, device='cuda:0')

    out = rope.apply_rotary_embeddings(x, x)
    assert out[0].shape == x.shape
    assert out[1].shape == x.shape


from cirilla.LLM_pieces import DynamicTanh, Dynamic_erf
from dataclasses import dataclass
import torch.nn as nn
from .modules import CirillaBaseModel
from .blocks import Decoder, DecoderArgs, InputEmbeddings
import torch
import torch.nn.functional as F

@dataclass
class MTPArgs(DecoderArgs):
    n_token_heads:int = 4
    vocab_size:int = 60_000
    tie_params:bool = False
    out_bias:bool = False

class CirillaMTP(
            nn.Module,
            CirillaBaseModel,
            pipeline_tag="text-generation",
            library_name="pytorch",
            license="mit"
    ):
    def __init__(self, args:MTPArgs=None):
        super().__init__()

        if isinstance(args, dict):
            args = MTPArgs(**args)

        if args is None:
            args = MTPArgs()

        self.args = args
        self._prepare_model()

    def _prepare_model(self):

        self.emb = InputEmbeddings(self.args)
        if self.args.layer_norm == "RMSNorm":
            self.layer_norm = nn.RMSNorm(self.args.dim)
        elif self.args.layer_norm == "Derf":
            self.layer_norm = Dynamic_erf(self.args.dim)
        elif self.args.layer_norm == "DyT":
            self.layer_norm = DynamicTanh(self.args.dim)
        else:
            raise ValueError(f"allowed layer norms: 'RMSNorm', 'Derf', 'DyT' ; got: {self.args.layer_norm}")
        self.decoder = Decoder(self.args)

        self.output = nn.Linear(self.args.dim, self.args.vocab_size, bias=self.args.out_bias)
        if self.args.tie_params:
            self.output.weight = self.emb.embeddings.weight

        token_args = {k:v for k,v in self.args.__dict__.items() if k in DecoderArgs.__dataclass_fields__}
        token_args['n_layers'] = 1
        self.token_head_args = DecoderArgs(**token_args)

        self.token_heads = [nn.Sequential(Decoder(self.token_head_args), type(self.layer_norm)(self.args.dim)) for _ in range(self.args.n_token_heads)]
        
        self.token_heads = nn.ModuleList(self.token_heads)

        self.n_params = sum(p.numel() for p in self.parameters() if p.requires_grad)

        self.to(self.args.device, dtype=self.args.dtype)
        
    def get_z(self, x) -> torch.Tensor:
        
        x = self.emb(x)

        if self.args.output_moe_weights:
            x, moe_weights = self.decoder(x)

            x = self.layer_norm(x)

            return x, moe_weights
        
        else:
            x = self.decoder(x)

            x = self.layer_norm(x)
        
            return x
        
    def get_heads(self, idx, z) -> torch.Tensor:
        return self.output(self.token_heads[idx](z))

    def forward(self, x) -> list[torch.Tensor]:
        x = self.get_z(x)
        return [head for head in self.get_heads(x)]

def mtp_training_step(self, data, pad_id) -> float:
    step_loss = 0.0
    n = 0

    torch.compiler.cudagraph_mark_step_begin()
    
    x = data[0]
    y = data[1]
    
    y_fill = torch.tensor([[pad_id] * self.model.args.n_token_heads] * y.shape[0], dtype=y.dtype, device=y.device)
    y = torch.hstack([y, y_fill])

    z = self.model.get_z(x)
    zd = z.detach()
    zd.requires_grad = True

    for i in range(self.model.args.n_token_heads):
        preds = self.model.get_heads(i, zd)
        loss = F.cross_entropy(
            preds.view(-1, self.model.args.vocab_size),
            y[:, i:-(self.model.args.n_token_heads - i)].reshape(-1),
            ignore_index=pad_id)
        step_loss += loss.item()
        n += 1
        loss.backward()

    z.backward(gradient=zd.grad)

    return step_loss / n

@torch.inference_mode()
def mtp_inference_step(self, data, pad_id) -> float:

    x = data[0]
    y = data[1]
    
    z = self.model.get_z(x)
    zd = z.detach()
    zd.requires_grad = True

    preds = self.model.get_heads(0, zd)
    loss = F.cross_entropy(
        preds.view(-1, self.model.args.vocab_size),
        y.reshape(-1),
        ignore_index=pad_id)

    return loss.item()


import torch
import torch.nn as nn
from dataclasses import dataclass
from megablocks import Arguments, MoE, dMoE
from .activations import Dynamic_erf, DynamicTanh

@dataclass
class MegablockArgs:
    num_experts: int = 4
    k: int = 2
    dim: int = 128
    d_ff: int = 256
    capacity_factor: float = 1.0
    impl: str = "grouped"   # or "sparse" Sparse MLP is not supported with triton >=3.2.0
    dtype_str:str = 'bfloat16'
    device:str = 'cuda'
    moe_zloss_weight:float = 0.1

    @property
    def dtype(self):
        return getattr(torch, self.dtype_str)

class MegablockMoE(nn.Module):
    def __init__(self, args:MegablockArgs):
        super().__init__()

        self.args = args

        if self.args.layer_norm == "RMSNorm":
            self.layer_norm = nn.RMSNorm(self.args.dim)
        elif self.args.layer_norm == "Derf":
            self.layer_norm = Dynamic_erf(self.args.dim)
        elif self.args.layer_norm == "DyT":
            self.layer_norm = DynamicTanh(self.args.dim)
        else:
            raise ValueError(f"allowed layer norms: 'RMSNorm', 'Derf', 'DyT' ; got: {self.args.layer_norm}")

        init_method = torch.nn.init.xavier_uniform_

        self.args = Arguments(
                hidden_size=args.dim,
                ffn_hidden_size=args.d_ff,
                moe_num_experts=args.num_experts,
                moe_capacity_factor=args.capacity_factor,
                moe_top_k=args.k,
                init_method=init_method,
                memory_optimized_mlp=True,
                mlp_type="mlp",
                mlp_impl=args.impl,
                fp16= args.dtype_str == 'float16',
                bf16= args.dtype_str == 'bfloat16',
                device=args.device,
                moe_zloss_weight=args.moe_zloss_weight
            )
        
        self.moe = MoE(
            self.args
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:

        x = self.layer_norm(x)
        # MegaBlocks expects (seq, batch, dim)
        x = x.transpose(0, 1).contiguous()

        x, _ = self.moe(x)
        del _

        x = x.transpose(0, 1)  # back to (batch, seq, dim)
        return (x,)

class MegablockdMoE(nn.Module):
    def __init__(self, args:MegablockArgs):
        super().__init__()

        self.args = args

        if self.args.layer_norm == "RMSNorm":
            self.layer_norm = nn.RMSNorm(self.args.dim)
        elif self.args.layer_norm == "Derf":
            self.layer_norm = Dynamic_erf(self.args.dim)
        elif self.args.layer_norm == "DyT":
            self.layer_norm = DynamicTanh(self.args.dim)
        else:
            raise ValueError(f"allowed layer norms: 'RMSNorm', 'Derf', 'DyT' ; got: {self.args.layer_norm}")

        init_method = torch.nn.init.xavier_uniform_
        
        self.args = Arguments(
                hidden_size=args.dim,
                ffn_hidden_size=args.d_ff,
                moe_num_experts=args.num_experts,
                moe_capacity_factor=args.capacity_factor,
                moe_top_k=args.k,
                init_method=init_method,
                memory_optimized_mlp=True,
                mlp_type="mlp",
                mlp_impl=args.impl,
                fp16= args.dtype_str == 'float16',
                bf16= args.dtype_str == 'bfloat16',
                device=args.device,
                moe_zloss_weight=args.moe_zloss_weight
            )
        
        self.moe = dMoE(
            self.args
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:

        x = self.layer_norm(x)
        # MegaBlocks expects (seq, batch, dim)
        x = x.transpose(0, 1).contiguous()

        x, _ = self.moe(x)
        del _

        x = x.transpose(0, 1)  # back to (batch, seq, dim)
        return (x,)



from cirilla.LLM_pieces.BERT_attention import BertAttention, BertAttentionArgs
from cirilla.LLM_pieces import RoPE
import pytest
import torch

@pytest.fixture
def get_rope():
    return RoPE(head_dim=128, seq_len=1024, device='cuda:0')

@pytest.mark.parametrize("n_heads, n_kv_heads, dim, soft_cap, device", [
    (4, 2, 128*4, 20, 'cuda:0'),
])
def test_bert_attention_initialization(n_heads, n_kv_heads, dim, soft_cap, device, get_rope):
    args = BertAttentionArgs(
        n_heads=n_heads,
        n_kv_heads=n_kv_heads,
        dim=dim,
        soft_cap=soft_cap,
        device=device
    )
    attention = BertAttention(args, get_rope).to(device, dtype=torch.bfloat16)
    
    # verify forward pass
    x = torch.randn(2, 512, dim, device=device, dtype=torch.bfloat16, requires_grad=True)
    out = attention(x)
    assert out.shape == (2, 512, dim)

    # very backward pass
    loss = out.sum()
    loss.backward()

    grads_ok = all(p.grad is not None for p in attention.parameters() if p.requires_grad)
    if not grads_ok:
        for name, p in attention.named_parameters():
            if p.requires_grad and p.grad is None:
                print(f"Parameter {name} has no gradient.")
        raise AssertionError("Not all parameters have gradients after initialization.")


import torch
import configparser
import os
from huggingface_hub import hf_hub_download
import json
import time
from safetensors.torch import load_file
from huggingface_hub import PyTorchModelHubMixin
from ..LLM_pieces import get_activation
from cirilla.LLM_pieces.sliding_window_attention import create_dynamic_block_mask

class CirillaBaseModel(PyTorchModelHubMixin):
    def __init__(self):
        pass
    
    def pull_model_from_hub(self, hf_repo_id:str, inference_mode:bool=False, map_device:str=None):
        model_args = self.args
        pulled_args = get_args_from_hub(hf_repo_id, type(self.args))

        if model_args != pulled_args:
            print(f"Current model args don't correspond to the HF model's args.\nCurrent args:\n{model_args}\nThe model will use the HF args:\n{pulled_args}")
            self.args = pulled_args
            if inference_mode:
                self.args.torch_compile = False
            if map_device is not None:
                self.args.device = map_device
            self._prepare_model()

        file_path = hf_hub_download(
            repo_id=hf_repo_id,
            filename="model.safetensors",
        )

        loaded = load_file(file_path)
        if "output.weight" not in loaded:
            loaded['output.weight'] = loaded["emb.embeddings.weight"]

        if inference_mode:
            new_state_dict = {}
            for key, value in loaded.items():
                if "_orig_mod." in key:
                    new_key = key.replace("_orig_mod.", "")
                    new_state_dict[new_key] = value
                else:
                    new_state_dict[key] = value

            if hasattr(self, 'decoder'):
                for att in self.decoder.attentions:
                    att.static_mask = False
                    att.mask = create_dynamic_block_mask

            self.load_state_dict(new_state_dict)

        else:
            self.load_state_dict(loaded)

    @staticmethod
    def mean_pooling(out, attention_mask):
        if attention_mask is None:
            return torch.mean(out, dim=1)
        
        mask_expanded = attention_mask.unsqueeze(-1).expand(out.size()).to(out.dtype)
        
        sum_embeddings = torch.sum(out * mask_expanded, dim=1)
        
        sum_mask = mask_expanded.sum(dim=1)
        
        return sum_embeddings / torch.clamp(sum_mask, min=1e-9)

def benchmark_model_part(model, x, label=""):
    model.train()
    x = x.contiguous()

    # Warmup (not measured)
    for _ in range(10):
        out = model(x)
        loss = out.sum()
        loss.backward()
        torch.cuda.synchronize()
        model.zero_grad(set_to_none=True)

    fwd_times, bwd_times = [], []
    fwd_mems, bwd_mems = [], []

    for _ in range(100):
        # Forward
        torch.cuda.synchronize()
        start_mem = torch.cuda.memory_allocated()
        start_time = time.time()

        out = model(x)
        loss = out.sum()

        torch.cuda.synchronize()
        fwd_times.append(time.time() - start_time)
        fwd_mems.append(torch.cuda.memory_allocated() - start_mem)

        # Backward
        torch.cuda.synchronize()
        start_mem = torch.cuda.memory_allocated()
        start_time = time.time()

        loss.backward()

        torch.cuda.synchronize()
        bwd_times.append(time.time() - start_time)
        bwd_mems.append(torch.cuda.memory_allocated() - start_mem)

        model.zero_grad(set_to_none=True)

    print(f"\n[{label}]")
    print(f"Forward time:   {sum(fwd_times)/len(fwd_times)*1000:.2f} ms")
    print(f"Backward time:  {sum(bwd_times)/len(bwd_times)*1000:.2f} ms")
    print(f"Forward memory: {sum(fwd_mems)/len(fwd_mems)/1024/1024:.2f} MB")
    print(f"Backward memory:{sum(bwd_mems)/len(bwd_mems)/1024/1024:.2f} MB")

def get_args_from_hub(hf_repo_id, args_type):
    file_path = hf_hub_download(
        repo_id=hf_repo_id,
        filename="config.json",
    )
    with open(file_path, "r") as f:
        config = json.load(f)
    args = args_type(**config[list(config.keys())[0]])

    return args

def select_torch_device():
    if torch.cuda.is_available():
        device = "cuda:0"
        # print(f"Using device: {device}")
        # print(f"Device name: {torch.cuda.get_device_name(device)}")
        # mem_gb = torch.cuda.get_device_properties(device).total_memory / 1024 ** 3
        # print(f"Device memory: {mem_gb:.2f} GB")
    elif getattr(torch, 'has_mps', False) and torch.backends.mps.is_available():
        device = "mps"
        # print("Using device: mps (Apple Metal Performance Shaders)")
    else:
        device = "cpu"
        # print("Using device: cpu")
        # print("NOTE: If you have a GPU, consider using it for training.")

    return device

def find_cache(start_dir='./'):
    for main_path, subfolders, files in os.walk(start_dir):
        if '.cirilla' in files:
            path_to_cache = os.path.join(main_path, '.cirilla')
            return path_to_cache

CACHE_PATH = None

def cache_or_fetch(category, variable, value=None):
    global CACHE_PATH
    if CACHE_PATH is None:
        CACHE_PATH = find_cache()
        if CACHE_PATH is None:
            CACHE_PATH = '.cirilla'

    config = configparser.ConfigParser()
    if os.path.exists(CACHE_PATH):
        config.read(CACHE_PATH)

    try:
        val = config[category][variable]
        if value is not None:
            config[category][variable] = str(value)
            with open(CACHE_PATH, 'w') as c:
                config.write(c)
        try:
            return int(val)
        except:
            return val
    except (KeyError, configparser.NoSectionError):
        if value is not None:
            if category not in config:
                config[category] = {}
            config[category][variable] = str(value)
            with open(CACHE_PATH, 'w') as c:
                config.write(c)
            try:
                return int(value)
            except:
                return value
        else:
            return None

def load_balancing_loss(expert_weights: torch.Tensor,
                        num_experts: int,
                        top_k: int,
                        eps: float = 1e-12) -> torch.Tensor:
    """
    Optimal load balancing loss for MoE (highly vectorized).

    Args:
        expert_weights: Tensor of shape [tokens, top_k] or [batch, seq, top_k],
                        containing the softmax gating weights for top-k experts.
        num_experts: Total number of experts (E).
        top_k: Number of experts per token.
        eps: Small epsilon for numerical stability.

    Returns:
        Scalar tensor: load balancing loss.
    """
    # Flatten to [tokens, top_k]
    if expert_weights.dim() == 3:
        expert_weights = expert_weights.reshape(-1, top_k)  # (B*S, k)

    # Convert top-k weights to dense expert distribution (E,)
    # Each token contributes its top-k weights to corresponding experts
    # Here we assume experts are indexed 0..num_experts-1 evenly across top-k
    # (this is equivalent to a soft one-hot expansion)
    # For optimal version, we sum contributions per expert
    # Generate a placeholder expert assignment for simplicity: evenly spread
    # For exact assignment, use topk_idx instead of this.
    # Here we assume top_k == num_experts for simplified demonstration
    if top_k != num_experts:
        # If top_k < num_experts, distribute weights evenly to simulate usage
        importance = expert_weights.sum(0) * (num_experts / top_k)
    else:
        importance = expert_weights.sum(0)  # (num_experts,)

    # Normalize to probability
    importance = importance / (importance.sum() + eps)

    # KL divergence to uniform distribution
    # KL(p || U) = sum_i p_i * log(p_i / u_i) = sum_i p_i * log(p_i) + log(num_experts)
    loss = (importance * (importance + eps).log()).sum()
    loss = loss + torch.log(torch.tensor(num_experts, device=expert_weights.device, dtype=expert_weights.dtype))

    return loss

def get_optims(model, use_muon_optim, optim, lr, weight_decay, moun_optim = get_activation("motif-technologies/optimizer")):

    if use_muon_optim:
        get_default_muon_param_groups = moun_optim.muon.get_default_muon_param_groups
        muon_param_groups = get_default_muon_param_groups(model)

        moptim = moun_optim.Muon(muon_param_groups, lr=lr, weight_decay=weight_decay)

        rest_of_params = [p for n, p in model.named_parameters() if n not in muon_param_groups[0]['names']]

        roptim = optim(rest_of_params, lr=lr, weight_decay=weight_decay, betas=(0.9, 0.95))

        return moptim, roptim
    
    else:
        return optim(model.parameters(), lr=lr, weight_decay=weight_decay, betas=(0.9, 0.95))



from .multi_turn_vllm import vllm_multi_turn
from .Ollama_curate import OllamaCurate
from .reason_gym_synthetic import get_synth_reasoning_dataset
from .rm_duplicate_instruct import rm_duplicate_instructs
from .witcher_mr_gather import multi_turn_gather
from .summaries_gather import gather_summaries, summaries_to_instruct

__all__ = [
    'vllm_multi_turn',
    'OllamaCurate',
    'get_synth_reasoning_dataset',
    'rm_duplicate_instructs',
    'multi_turn_gather',
    'gather_summaries',
    'summaries_to_instruct'
]


import torch.nn as nn
import torch

class RoPE(nn.Module):
    def __init__(self, head_dim: int, seq_len: int, device="cuda", theta: float = 10000.0, dtype=torch.bfloat16):
        super().__init__()
        assert head_dim % 2 == 0, "head_dim must be even"
        self.dtype = dtype

        theta_numerator = torch.arange(0, head_dim, 2, device=device, dtype=torch.float32)
        inv_freq = 1.0 / (theta ** (theta_numerator / head_dim))  # (head_dim/2)
        t = torch.arange(seq_len, device=device, dtype=torch.float32)  # (seq_len)
        freqs = torch.outer(t, inv_freq)  # (seq_len, head_dim/2)

        # (1, seq_len, 1, head_dim/2)
        cos = torch.cos(freqs)[None, :, None, :].to(dtype)
        sin = torch.sin(freqs)[None, :, None, :].to(dtype)

        self.register_buffer("cos", cos, persistent=False)
        self.register_buffer("sin", sin, persistent=False)

    def apply_rotary_embeddings(self, xq: torch.Tensor, xk: torch.Tensor) -> torch.Tensor:
        seq_len = xq.size(1)

        cos = self.cos[:, :seq_len, :, :]  # [1, seq_len, 1, head_dim/2]
        sin = self.sin[:, :seq_len, :, :]  # same

        # Split last dim into even/odd
        xq_even, xq_odd = xq[..., ::2], xq[..., 1::2]
        xk_even, xk_odd = xk[..., ::2], xk[..., 1::2]

        xq_rot = torch.stack([xq_even * cos - xq_odd * sin,
                              xq_even * sin + xq_odd * cos], dim=-1)
        xk_rot = torch.stack([xk_even * cos - xk_odd * sin,
                              xk_even * sin + xk_odd * cos], dim=-1)

        xq_out = xq_rot.flatten(-2)
        xk_out = xk_rot.flatten(-2)

        return xq_out, xk_out



from fuzzywuzzy import fuzz
import os
import re
import json
import numpy as np
from pathlib import Path
from typing import Union

def rm_duplicate_instructs(main_dir: Union[Path, list[Path]], save_to: Path):
    files = []

    for main_path, subfolders, _files in os.walk(main_dir):
        for _file in _files:
            if _file.endswith(".json"):
                files.append(os.path.join(main_path, _file))

    print(f"Found {len(files)} files.")

    all_names = set(
        re.sub(r'\d+', '', os.path.basename(file).replace('_', ''))
        for file in files
    )

    valid_paths = []

    for name in all_names:
        same_files = [
            file for file in files
            if re.sub(r'\d+', '', os.path.basename(file).replace('_', '')) == name
        ]
        
        data = []
        for file in same_files:
            with open(file, 'r') as f:
                data.append(json.load(f))

        n = len(data)
        sim_matrix = np.zeros((n, n))

        for i in range(n):
            for j in range(n):
                text_i = data[i].get('question', '') + ' ' + data[i].get('answer', '')
                text_j = data[j].get('question', '') + ' ' + data[j].get('answer', '')
                sim_matrix[i, j] = fuzz.ratio(text_i, text_j)

        ids_delete = set()
        for i in range(n):
            for j in range(i + 1, n):
                if sim_matrix[i, j] > 90:
                    ids_delete.add(j)

        for i in range(n):
            if i not in ids_delete:
                valid_paths.append(same_files[i])

    print(f"Remaining files: {len(valid_paths)}")

    data_to_save = []

    for name in all_names:
        same_files = [
            file for file in valid_paths
            if re.sub(r'\d+', '', os.path.basename(file).replace('_', '')) == name
        ]
        for file in same_files:
            model = file.split('/')[-2]
            qa = json.load(open(file, 'r'))
            if (qa['question'] != '') and (qa['answer'] != ''):
                data_ = ({
                    'subject': name.split('.')[0],
                    'text': [{'role':'user', 'content': qa['question']}, {'role': 'assistant', 'content': qa['answer']}],
                    'data type': 'conv',
                    'model': model,
                })

                data_to_save.append(data_)

    with open(save_to, 'w') as f:
        for d in data_to_save:
            f.write(json.dumps(d) + '\n')



import torch
import torch.nn as nn
import torch.nn.functional as F
from dataclasses import dataclass
from .activations import Dynamic_erf, DynamicTanh


@dataclass
class SwiGLUArgs:
    dim:int=128
    d_ff:int=256 # hidden dim
    assert d_ff % 2 == 0
    drop:float=0.1

class SwiGLU(nn.Module):
    def __init__(self, args: SwiGLUArgs):
        super().__init__()
        self.dim = args.dim
        self.d_ff = args.d_ff

        self.w1a = nn.Linear(args.dim, args.d_ff)
        self.w1b = nn.Linear(args.dim, args.d_ff)
        self.w2 = nn.Linear(args.d_ff, args.dim)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        a = self.w1a(x)
        b = self.w1b(x)

        x = F.silu(a) * b

        x = self.w2(x)
        return x

@dataclass
class SMoEArgs:
    num_experts:int=8
    k:int=2
    dim:int=128
    dtype_str:str = 'bfloat16'
    device:str = 'cuda'
    d_ff:int=256 # hidden dim
    layer_norm:str = "RMSNorm"

    @property
    def dtype(self):
        return getattr(torch, self.dtype_str)

class SMoE(nn.Module):
    def __init__(self, args:SMoEArgs, experts:list[SwiGLU]):
        super().__init__()
        self.n_experts = args.num_experts
        self.k = args.k
        self.gating = nn.Linear(args.dim, args.num_experts)
        self.experts = nn.ModuleList(experts)
        self.args = args

        if self.args.layer_norm == "RMSNorm":
            self.layer_norm = nn.RMSNorm(self.args.dim)
        elif self.args.layer_norm == "Derf":
            self.layer_norm = Dynamic_erf(self.args.dim)
        elif self.args.layer_norm == "DyT":
            self.layer_norm = DynamicTanh(self.args.dim)
        else:
            raise ValueError(f"allowed layer norms: 'RMSNorm', 'Derf', 'DyT' ; got: {self.args.layer_norm}")

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.layer_norm(x)                           # (B,S,D)

        logits = self.gating(x)                       # (B,S,E)
        topk_vals, topk_idx = torch.topk(logits, self.k, dim=-1)

        topk_w = F.softmax(topk_vals, dim=-1)        # (B,S,k)
        one_hot = F.one_hot(topk_idx, num_classes=self.n_experts).to(x.dtype)

        weights_per_expert = (one_hot * topk_w.unsqueeze(-1)).sum(dim=2)  # (B,S,E)

        out = torch.zeros_like(x)
        for ex_idx, expert in enumerate(self.experts):
            w = weights_per_expert[..., ex_idx].unsqueeze(-1)  # (B,S,1)
            out = out + w * expert(x)                         # expert(x) -> (B,S,D)
        return out, weights_per_expert



from torch.nn.attention.flex_attention import create_block_mask, flex_attention, BlockMask
from functools import lru_cache, partial
from .RoPE import RoPE
import torch.nn as nn
from dataclasses import dataclass
from typing import Union
import torch
from .activations import Dynamic_erf, DynamicTanh

SLIDING_WINDOW = 512

def sliding_window_causal(b, h, q_idx, kv_idx) -> torch.Tensor:
    causal_mask = q_idx >= kv_idx
    window_mask = q_idx - kv_idx <= SLIDING_WINDOW
    return causal_mask & window_mask

def full_attention(b, h, q_idx, kv_idx) -> torch.Tensor:
    return torch.tensor(True, device=q_idx.device)

def create_static_block_mask(sliding_window_causal, q_len, kv_len, device='cuda', window_size=512) -> BlockMask:
    global SLIDING_WINDOW
    SLIDING_WINDOW = window_size
    # B,H set to None means that the mask is broadcasted for those dimentions as it doesn't require any calculation anyway
    return create_block_mask(sliding_window_causal, B=None, H=None, Q_LEN=q_len, KV_LEN=kv_len, device=device)

@lru_cache(maxsize=32)
def create_dynamic_block_mask(sliding_window_causal, q_len=2048, kv_len=2048, device='cuda', window_size=512) -> BlockMask:
    global SLIDING_WINDOW
    SLIDING_WINDOW = window_size
    # B,H set to None means that the mask is broadcasted for those dimentions as it doesn't require any calculation anyway
    return create_block_mask(sliding_window_causal, B=None, H=None, Q_LEN=q_len, KV_LEN=kv_len, device=device)

@dataclass
class AttentionArgs:
    n_heads:int = 16
    n_kv_heads:int = 4
    dim:int = 128*16
    static_mask:bool = True
    window_size:int = 512
    device:str = 'cuda:0'
    layer_norm:str = "RMSNorm"

class SlidingWindowAttention(nn.Module):
    def __init__(self, args: AttentionArgs, rope:RoPE, mask:Union[BlockMask, create_dynamic_block_mask]=None, score_mod:callable=None):
        super().__init__()

        self.args = args

        self.n_kv_heads = args.n_heads if args.n_kv_heads is None else args.n_kv_heads
        self.n_heads_q = args.n_heads
        self.n_rep = self.n_heads_q // self.n_kv_heads
        self.head_dim = args.dim // args.n_heads
        self.static_mask = args.static_mask
        
        if self.args.layer_norm == "RMSNorm":
            self.layer_norm = nn.RMSNorm(self.args.dim)
        elif self.args.layer_norm == "Derf":
            self.layer_norm = Dynamic_erf(self.args.dim)
        elif self.args.layer_norm == "DyT":
            self.layer_norm = DynamicTanh(self.args.dim)
        else:
            raise ValueError(f"allowed layer norms: 'RMSNorm', 'Derf', 'DyT' ; got: {self.args.layer_norm}")

        self.wo = nn.Linear(args.n_heads * self.head_dim, args.dim, bias=False)

        self.wq = nn.Linear(args.dim, args.n_heads * self.head_dim, bias=False)
        self.wk = nn.Linear(args.dim, args.n_kv_heads * self.head_dim, bias=False)
        self.wv = nn.Linear(args.dim, args.n_kv_heads * self.head_dim, bias=False)
        
        self.hq_dim = self.n_heads_q * self.head_dim
        self.hkv_dim = self.n_kv_heads * self.head_dim

        self.rope = rope
        self.mask = mask if not isinstance(mask, BlockMask) else None
        self.score_mode = score_mod
        self.window_size = args.window_size

        self.attn = partial(flex_attention, block_mask=mask if mask is not None and isinstance(mask, BlockMask) else None,
                            score_mod=score_mod if score_mod is not None else None, enable_gqa= self.n_heads_q != self.n_kv_heads)\
                        if self.static_mask else None

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        batch_size, seq_len, dim = x.shape

        x = self.layer_norm(x)

        xq = self.wq(x)
        xk = self.wk(x)
        xv = self.wv(x)

        xq = xq.view(batch_size, seq_len, self.n_heads_q, self.head_dim)
        xk = xk.view(batch_size, seq_len, self.n_kv_heads, self.head_dim)
        xv = xv.view(batch_size, seq_len, self.n_kv_heads, self.head_dim)

        xq, xk = self.rope.apply_rotary_embeddings(xq, xk)

        # (b, seq, h_q, head_dim) -> (b, h_q, seq, head_dim)
        xq = xq.transpose(1,2)
        xk = xk.transpose(1,2)
        xv = xv.transpose(1,2)

        if self.static_mask:
            out = self.attn(xq, xk, xv)
        else:
            mask = self.mask(sliding_window_causal, xq.shape[2], xk.shape[2], device=xq.device, window_size=self.window_size)
            out = flex_attention(xq, xk, xv, block_mask=mask, score_mod=self.score_mode, enable_gqa= self.n_heads_q != self.n_kv_heads)
        
        out = out.transpose(1,2).contiguous().view(batch_size, seq_len, dim) # (b, seq, dim)
        return self.wo(out) #(b, seq, dim)
    
    @torch.inference_mode()
    def forward_with_cache(self, x: torch.Tensor, cur_pos:int, max_batch:int=1, chunked_prefill:bool=False, non_finished_ids:torch.Tensor=None) -> torch.Tensor:

        batch_size, seq_len, dim = x.shape

        if not hasattr(self, 'k_cache'):
            self.k_cache = torch.zeros(max_batch,
                                        self.args.context_window,
                                        self.n_kv_heads,
                                        self.head_dim,
                                        device=self.args.device,
                                        dtype=x.dtype)
            
            self.v_cache = torch.zeros(max_batch,
                                        self.args.context_window,
                                        self.n_kv_heads,
                                        self.head_dim,
                                        device=self.args.device,
                                        dtype=x.dtype)
            
        if non_finished_ids is None:
            non_finished_ids = torch.arange(self.k_cache.size(0), device=x.device)
        
        x = self.layer_norm(x)

        xq = self.wq(x)
        xk = self.wk(x)
        xv = self.wv(x)

        xq = xq.view(batch_size, seq_len, self.n_heads_q, self.head_dim)
        xk = xk.view(batch_size, seq_len, self.n_kv_heads, self.head_dim)
        xv = xv.view(batch_size, seq_len, self.n_kv_heads, self.head_dim)
        
        cos = self.rope.cos[:, cur_pos:cur_pos+seq_len, :, :]
        sin = self.rope.sin[:, cur_pos:cur_pos+seq_len, :, :]

        xq_even, xq_odd = xq[..., ::2], xq[..., 1::2]
        xk_even, xk_odd = xk[..., ::2], xk[..., 1::2]

        xq_rot = torch.stack([xq_even * cos - xq_odd * sin,
                              xq_even * sin + xq_odd * cos], dim=-1)
        xk_rot = torch.stack([xk_even * cos - xk_odd * sin,
                              xk_even * sin + xk_odd * cos], dim=-1)

        xq = xq_rot.flatten(-2)
        xk = xk_rot.flatten(-2)

        self.k_cache[non_finished_ids, cur_pos:cur_pos+seq_len, :, :] = xk
        self.v_cache[non_finished_ids, cur_pos:cur_pos+seq_len, :, :] = xv

        window_end = cur_pos + seq_len

        if chunked_prefill:
            window_start = max(0, window_end - self.window_size - seq_len)
            xk = self.k_cache[non_finished_ids, window_start:window_end] # chunked prefill uses an overlapping window atten~(window, 2*window)
            xv = self.v_cache[non_finished_ids, window_start:window_end]

        else:
            window_start = max(0, window_end - self.window_size)
            xk = self.k_cache[non_finished_ids, window_start:window_end] # rolling buffer cache
            xv = self.v_cache[non_finished_ids, window_start:window_end]

        # (b, seq, h_q, head_dim) -> (b, h_q, seq, head_dim)
        xq = xq.transpose(1,2)
        xk = xk.transpose(1,2)
        xv = xv.transpose(1,2)

        if not chunked_prefill: # the kv size is limited, so this is the same as sliding window causal mask
            attention_mask = full_attention
        
        elif chunked_prefill and xk.size(2) <= self.window_size:
            attention_mask = sliding_window_causal

        elif chunked_prefill and xk.size(2) > self.window_size:
            
            def offset_sliding_window_causal(b, h, q_idx, kv_idx):
                q_idx += seq_len
                causal_mask = q_idx >= kv_idx
                window_mask = q_idx - kv_idx <= SLIDING_WINDOW
                return causal_mask & window_mask

            attention_mask = offset_sliding_window_causal

        mask = create_dynamic_block_mask(attention_mask,
                                        q_len=seq_len,
                                        kv_len=xk.size(2),
                                        device=xq.device,
                                        window_size=self.window_size)

        out = flex_attention(xq, xk, xv,
                            block_mask=mask,
                            score_mod=self.score_mode,
                            enable_gqa=\
                                self.n_heads_q != self.n_kv_heads
                            )

        out = out.transpose(1,2).contiguous().view(batch_size, seq_len, dim) # (b, seq, dim)
        return self.wo(out) #(b, seq, dim)
    
    def _clear_cache(self):
        if hasattr(self, 'k_cache'):
            del self.k_cache, self.v_cache



from attn_gym.utils import visualize_attention_scores
import attn_gym.mods
import attn_gym.masks
import attn_gym.paged_attention

__all__ = ["visualize_attention_scores", "mods", "masks", "paged_attention"]



try:
    from cirilla.LLM_pieces import SMoE, MegablockdMoE, MegablockMoE
    from cirilla.LLM_pieces.SMoE import MegablockArgs
except ImportError:
    test_megablocks = False

from cirilla.LLM_pieces.SMoE import SwiGLUArgs, SwiGLU, SMoEArgs
import torch

def test_expert():
    expert_args = SwiGLUArgs()
    expert = SwiGLU(expert_args).to('cuda')
    x = torch.randn(4, 10, expert_args.dim, device='cuda')
    out = expert(x)
    assert out.shape == (4, 10, expert_args.dim)

    loss = out.sum()
    loss.backward()

    grads_ok = all(p.grad is not None for p in expert.parameters() if p.requires_grad)
    if not grads_ok:
        for name, p in expert.named_parameters():
            if p.requires_grad and p.grad is None:
                print(f"Parameter {name} has no gradient.")
        raise AssertionError("Not all parameters have gradients after initialization.")
    
def test_pytorch_smoe():
    smoeargs = SMoEArgs(num_experts=4, k=2, dim=128, d_ff=256, device='cuda')
    experts = [SwiGLU(SwiGLUArgs(dim=128, d_ff=256)).to('cuda') for _ in range(smoeargs.num_experts)]
    model = SMoE(smoeargs, experts).to('cuda')

    x = torch.randn(4, 10, smoeargs.dim, device='cuda')
    out = model(x)
    assert isinstance(out, tuple)
    assert isinstance(out[0], torch.Tensor)
    assert isinstance(out[1], torch.Tensor)

    out, weights = out
    assert out.shape == (4, 10, smoeargs.dim)
    assert weights.shape == (4, 10, smoeargs.num_experts)

    loss = out.sum()
    loss.backward()

    grads_ok = all(p.grad is not None for p in model.parameters() if p.requires_grad)
    if not grads_ok:
        for name, p in model.named_parameters():
            if p.requires_grad and p.grad is None:
                print(f"Parameter {name} has no gradient.")
        raise AssertionError("Not all parameters have gradients after initialization.")

if test_megablocks:
    def test_megablock_moe():
        megablock_args = MegablockArgs(num_experts=4, k=2, dim=128, d_ff=256, device='cuda')
        model = MegablockMoE(megablock_args).to('cuda')

        x = torch.randn(4, 10, megablock_args.dim, device='cuda', dtype=torch.bfloat16)
        out = model(x)
        assert isinstance(out, tuple)
        assert isinstance(out[0], torch.Tensor)

        out = out[0]
        assert out.shape == (4, 10, megablock_args.dim)

        loss = out.sum()
        loss.backward()

        grads_ok = all(p.grad is not None for n, p in model.named_parameters() if p.requires_grad if n != "moe.experts.bias")
        if not grads_ok:
            for name, p in model.named_parameters():
                if p.requires_grad and p.grad is None:
                    print(f"Parameter {name} has no gradient.")
            raise AssertionError("Not all parameters have gradients after initialization.")

    def test_megablock_dmoe():
        megablock_args = MegablockArgs(num_experts=4, k=2, dim=128, d_ff=256, device='cuda')
        model = MegablockdMoE(megablock_args).to('cuda')

        x = torch.randn(4, 10, megablock_args.dim, device='cuda', dtype=torch.bfloat16)
        out = model(x)
        assert isinstance(out, tuple)
        assert isinstance(out[0], torch.Tensor)

        out = out[0]
        assert out.shape == (4, 10, megablock_args.dim)

        loss = out.sum()
        loss.backward()

        grads_ok = all(p.grad is not None for n, p in model.named_parameters() if p.requires_grad if n != "moe.experts.bias")
        if not grads_ok:
            for name, p in model.named_parameters():
                if p.requires_grad and p.grad is None:
                    print(f"Parameter {name} has no gradient.")
            raise AssertionError("Not all parameters have gradients after initialization.")


from pathlib import Path
from .modules import cache_or_fetch
import torch
import json
import random
from torch.utils.data import IterableDataset
from typing import Union
from .tokenizer_modules import CirillaTokenizer

KEYBOARD_NEIGHBORS = {
    'a': 'qs', 'b': 'vn', 'c': 'xv', 'd': 'sf', 'e': 'wr', 'f': 'dg', 'g': 'fh',
    'h': 'gj', 'i': 'uo', 'j': 'hk', 'k': 'jl', 'l': 'k', 'm': 'n', 'n': 'bm',
    'o': 'ip', 'p': 'o', 'q': 'wa', 'r': 'et', 's': 'ad', 't': 'ry', 'u': 'iy',
    'v': 'cb', 'w': 'qe', 'x': 'zc', 'y': 'tu', 'z': 'x'
    }

class GenericDataset:
    def __init__(self, path:Union[Path, tuple[Path]]='./training_dataset.jsonl',
                shuffle_path=False,
                device:torch.device='cuda',
                tokenizer:CirillaTokenizer=None,
                max_len:int=32,
                pad_token:str='<pad>',
                eos_token:str='<eos>',
                sos_token:str='<sos>',
                user_token:str='<|user|>',
                suffix_tokens:list[str]=None, #['<cls>']
                prefix_tokens:list[str]=None,
                random_spelling_mistake_prob:float=0.,
                random_missing_char_prob:float=0.
                ):
        
        self.path = path
        self.shuffle_path = shuffle_path
        self.device = device
        self.tokenizer = tokenizer
        self.max_len = max_len
        self.suffix_tokens = suffix_tokens
        self.prefix_tokens = prefix_tokens
        self.random_spelling_mistake_prob = random_spelling_mistake_prob
        self.random_missing_char_prob = random_missing_char_prob

        if self.tokenizer is not None:
            self.sos_token = sos_token
            self.user_token_id = tokenizer.convert_tokens_to_ids(user_token)
            self.pad_token_id = tokenizer.convert_tokens_to_ids(pad_token)
            self.eos_token_id = tokenizer.convert_tokens_to_ids(eos_token)

        if isinstance(self.path, list):
            self.path = tuple(self.path)

        if isinstance(self.path, str):
            self.path = (self.path,)

        for p in self.path:

            if cache_or_fetch('DATA_LEN', p) is None:
                with open(p, 'r', encoding='utf-8') as f:
                    count = sum(1 for _ in f)
                cache_or_fetch('DATA_LEN', p, count)

            if cache_or_fetch('SHUFFLED', p) is None and shuffle_path:
                with open(p, 'r') as f:
                    data = [json.loads(line) for line in f]
                
                random.shuffle(data)
                with open(p, 'w') as f:
                    for d in data:
                        f.write(json.dumps(d) + '\n')

                del data

            cache_or_fetch('SHUFFLED', p, 1)
        
        self.path_signature = '-'.join(self.path)
        cache_or_fetch('DATA_LEN', self.path_signature, sum(cache_or_fetch('DATA_LEN', p) for p in self.path))

    def __len__(self):
        return cache_or_fetch('DATA_LEN', self.path_signature)
    
    def _apply_random_spelling_mistake(self, text):
        new_text = []
        for letter in text:
            if not letter.isupper():

                if random.random() < self.random_missing_char_prob: # skip letter
                    continue

                elif random.random() < self.random_spelling_mistake_prob:
                    letter_lower = letter.lower()
                    if letter_lower in KEYBOARD_NEIGHBORS:
                        new_letter = random.choice(KEYBOARD_NEIGHBORS[letter_lower])
                    else:
                        new_letter = letter
                    new_text.append(new_letter)
                else:
                    new_text.append(letter)
            else:
                new_text.append(letter)
        return ''.join(new_text)

class JSONLDataset(IterableDataset, GenericDataset):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    
    def __iter__(self):

        for p in self.path:
            with open(p, 'r') as f:

                for line in f:
                    line = json.loads(line)

                    match line['data type']:

                        case 'plain text':

                            if self.tokenizer is not None:

                                if self.random_spelling_mistake_prob > 0. or self.random_missing_char_prob > 0.:

                                    v = self.tokenizer(self.sos_token + line['text'], return_tensors='pt', padding='do_not_pad',
                                                                    truncation=True, max_length=self.max_len+1)
                                    
                                    out_tokens = v['input_ids'].squeeze(0)
                                    token_shape = out_tokens.shape[0]

                                    if token_shape <= self.max_len:
                                        out_tokens = torch.concat([out_tokens, torch.tensor([self.eos_token_id])] + \
                                                        [torch.tensor([self.pad_token_id])] * (self.max_len - token_shape), dim=0)
                                    
                                    out_tokens = out_tokens.to(self.device)
                                    
                                    line['text'] = self._apply_random_spelling_mistake(line['text'])
                                    
                                    tokenized_data = self.tokenizer(self.sos_token + line['text'], return_tensors='pt', padding='do_not_pad',
                                                                    truncation=True, max_length=self.max_len+1)
                                    
                                    in_tokens = tokenized_data['input_ids'].squeeze(0)
                                    token_shape = in_tokens.shape[0]

                                    if token_shape <= self.max_len:
                                        in_tokens = torch.concat([in_tokens, torch.tensor([self.eos_token_id])] + \
                                                        [torch.tensor([self.pad_token_id])] * (self.max_len - token_shape), dim=0)
                                    
                                    in_tokens = in_tokens.to(self.device)

                                    yield in_tokens[:-1], out_tokens[1:]

                                else:

                                    if self.random_spelling_mistake_prob > 0. or self.random_missing_char_prob > 0.:
                                        line['text'] = self._apply_random_spelling_mistake(line['text'])
                                        
                                    tokenized_data =  self.tokenizer(self.sos_token + line['text'], return_tensors='pt', padding='do_not_pad',
                                                                    truncation=True, max_length=self.max_len+1)
                                    
                                    tokens = tokenized_data['input_ids'].squeeze(0)
                                    # mask = tokenized_data['attention_mask'].squeeze(0)
                                    token_shape = tokens.shape[0]

                                    # mask = torch.concat([mask, torch.ones(1, dtype=torch.int64), torch.zeros(self.max_len - token_shape, dtype=torch.int64)],
                                    #                 dim=0).to(self.device)
                                    if token_shape <= self.max_len:
                                        tokens = torch.concat([tokens, torch.tensor([self.eos_token_id])] + \
                                                        [torch.tensor([self.pad_token_id])] * (self.max_len - token_shape), dim=0)
                                    
                                    tokens = tokens.to(self.device)
                                    yield tokens[:-1], tokens[1:]#, mask[:-1]

                            else:
                                yield line['text']

                        case 'conv':

                            if self.tokenizer is not None:
                                tokens =  self.tokenizer.apply_chat_template(line['text'], return_tensors='pt', padding='do_not_pad',
                                                                                    truncation=True, max_length=self.max_len+1, add_generation_prompt=False)
                                tokens = tokens.squeeze(0).to(self.device)
                                tokens_shape = tokens.shape[0]

                                # mask = torch.zeros(self.max_len, dtype=torch.int64, device=self.device)
                                # mask[:tokens_shape] = 1
                                
                                if tokens_shape <= self.max_len :
                                    tokens = torch.concat(
                                        [tokens, torch.tensor([self.user_token_id], device=self.device),] + \
                                        [torch.tensor([self.pad_token_id], device=self.device)] * (self.max_len - tokens_shape))
                                    
                                    # mask[tokens_shape] = 1
                                
                                yield tokens[:-1], tokens[1:]#, mask
                            else:
                                yield '\n'.join([l['content'] for l in line['text']])

                        case 'bert':

                            if self.tokenizer is not None:
                                text = line['text']
                                
                                if self.prefix_tokens is not None:
                                    text = "".join(self.prefix_tokens) + text

                                if self.suffix_tokens is not None:
                                    text += "".join(self.suffix_tokens)

                                tokenized_data =  self.tokenizer(text, return_tensors='pt', padding='max_length',
                                                                truncation=True, max_length=self.max_len)
                                
                                tokens = tokenized_data['input_ids'].squeeze(0).to(self.device)
                                mask = tokenized_data['attention_mask'].squeeze(0).to(self.device)
                                yield tokens, mask, torch.tensor(line['label'], dtype=torch.int64 if not isinstance(line['label'], list) else torch.bfloat16, device=self.device)
                            else:
                                yield line['text']

                        case _:
                            raise NotImplementedError(f"Data type {line['data type']} is not supported, use one of: ['plain text', 'conv', 'bert']")



from huggingface_hub import HfApi, PyTorchModelHubMixin, hf_hub_download
from huggingface_hub.repocard import metadata_eval_result, metadata_save
import json
import tempfile
from pathlib import Path
from .model import Cirilla
from dataclasses import dataclass
import torch

def push_model_to_hub(repo_id,
                model:Cirilla,
                loss:float,
                dataset_name:str,
                private:bool=False,
                optmizer_states_path:Path=None,
                tags:list[str]=["pytorch", "text-generation", "moe", "custom_code"],
                license:str='mit',
                languages:list[str] = ['en'],
                model_card:str = None
                ):

  repo_name = repo_id
  api = HfApi()

  repo_url = api.create_repo(
        repo_id=repo_id,
        private=private,
        exist_ok=True,
  )

  with tempfile.TemporaryDirectory() as tmpdirname:
    local_directory = Path(tmpdirname)

    metadata = {}
    metadata["tags"] = tags
    metadata['library'] = 'pytorch'
    metadata['license'] = license
    metadata['language'] = languages

    eval = metadata_eval_result(
        model_pretty_name=repo_name,
        task_pretty_name="text-generation",
        task_id="text-generation",
        metrics_pretty_name="Cross Entropy Loss",
        metrics_id="CEL",
        metrics_value=f'{loss:.3f}',
        dataset_pretty_name=dataset_name,
        dataset_id=dataset_name,
      )

    metadata = {**metadata, **eval}

    if model_card is None:
        model_card = f"""
# Random Pytorch model used as a demo to show how to push custom models to HF hub
| parameters | precision |
| :--------: | :-------: |
|{((model.n_params/1e6) if hasattr(model, 'n_params') else int(sum(p.numel() / 1e6 for p in model.parameters()))):.2f} M|{'BF16' if model.args.dtype == torch.bfloat16 else 'FP32'}|
"""

    readme_path = local_directory / "README.md"
    readme = ""
    if readme_path.exists():
        with readme_path.open("r", encoding="utf8") as f:
          readme = f.read()
    else:
      readme = model_card

    with readme_path.open("w", encoding="utf-8") as f:
      f.write(readme)

    if optmizer_states_path is not None:
        local_states = local_directory / "optimizer_states.pt"
        states = torch.load(optmizer_states_path, map_location='cpu')
        torch.save(states, local_states)
          

    model.push_to_hub(repo_id)

    metadata_save(readme_path, metadata)

    api.upload_folder(
          repo_id=repo_id,
          folder_path=local_directory,
          path_in_repo=".",
    )

    # print(f"Your model is pushed to the Hub. You can view your model here: {repo_url}")

if __name__ == "__main__":

    from dataclasses import dataclass
    import torch
    import torch.nn as nn
    import torch.optim as optim

    @dataclass
    class hypers:
        in_size: int = 64
        out_size: int = 1
        hidden_size: int = 16
        epochs: int = 3
        batch_size: int = 16
        lr: float = 1e-3
        dtype_str:str = 'bfloat16'

        @property
        def dtype(self):
            return getattr(torch, self.dtype_str)

    class NN(
        nn.Module,
        PyTorchModelHubMixin,
        pipeline_tag="text-generation",
        library_name="pytorch",
        license="mit"
    ):
        def __init__(self, args:dataclass):
            super().__init__()
            if isinstance(args, dict):
               args = hypers(**args)
            self.l1 = nn.Linear(args.in_size, args.hidden_size)
            self.l2 = nn.Linear(args.hidden_size, args.out_size)
            self.to(dtype=args.dtype)

        def forward(self, x):
            x = self.l1(x)
            x = torch.relu(x)
            x = self.l2(x)
            return x

    args = hypers()
    
    print(args.dtype)

    dataset = torch.utils.data.TensorDataset(torch.rand(1024, args.in_size), torch.rand(1024, args.out_size))
    dataloader = torch.utils.data.DataLoader(dataset, shuffle=True, batch_size=16)

    model = NN(args)
    optimizer = optim.SGD(model.parameters(), lr=args.lr)
    criterion = nn.MSELoss()

    for x, y in dataloader:
        x, y = x.to(dtype=args.dtype), y.to(dtype=args.dtype)
        pred = model(x)
        loss = criterion(pred, y)

        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

    push_model_to_hub("AnthonyPa57/HF-torch-demo2", model, loss, 'pretraining', save_locally='./test_model')
    
    repo_id = "AnthonyPa57/HF-torch-demo2"
    filename = "config.json"

    file_path = hf_hub_download(
        repo_id=repo_id,
        filename=filename,
    )

    with open(file_path, "r") as f:
        config = json.load(f)

    args = hypers(**config[list(config.keys())[0]])
    print(args)

    model_hf = NN(args)
    model_hf.from_pretrained("AnthonyPa57/HF-torch-demo2")

    print(model_hf)

    with open('./test_model/config.json', 'r') as f:
        config = json.load(f)
    args = hypers(**config[list(config.keys())[0]])
    model_hf = NN(args)
    model_hf.from_pretrained('./test_model')
    print(model_hf)


from reasoning_gym.composite import DatasetSpec
import reasoning_gym
import json
import random
from pathlib import Path

specs = [
    DatasetSpec(name='aiw', weight=2, config={"max_entities":10}),
    DatasetSpec(name='basic_arithmetic', weight=2, config={"whitespace":"random"}),
    DatasetSpec(name='caesar_cipher', weight=2, config={}),
    DatasetSpec(name='calendar_arithmetic', weight=2, config={}),
    DatasetSpec(name='chain_sum', weight=2, config={}),
    DatasetSpec(name='decimal_chain_sum', weight=2, config={}),
    DatasetSpec(name='family_relationships', weight=2, config={}),
    DatasetSpec(name='fraction_simplification', weight=2, config={"styles":["plain"]}),
    # DatasetSpec(name='gsm_symbolic', weight=1, config={"difficulty":1.0}), # <- doesnt work: Object of type Fraction is not JSON serializable
    DatasetSpec(name='knights_knaves', weight=2, config={}),
    DatasetSpec(name='leg_counting', weight=2, config={}),
    DatasetSpec(name='letter_counting', weight=2, config={}),
    DatasetSpec(name='needle_haystack', weight=2, config={"min_num_statements":2, "max_num_statements":10}),
    DatasetSpec(name='number_filtering', weight=2, config={"max_numbers":5}),
    DatasetSpec(name='number_sequence', weight=2, config={"min_terms":4, "max_terms":8}),
    DatasetSpec(name='sentence_reordering', weight=2, config={}),
    DatasetSpec(name='simple_equations', weight=2, config={"min_terms":1, "max_terms":3}),
    DatasetSpec(name='simple_geometry', weight=1, config={}),
    DatasetSpec(name='spell_backward', weight=2, config={}),
    DatasetSpec(name='syllogism', weight=2, config={}),
]

def get_synth_reasoning_dataset(out_path: Path, n_samples: int, specs:list[DatasetSpec]=specs):
    """
    create synthetic reasoning dataset with reasoning_gym

    Args:
        out_path (Path): path to save the synthetic reasoning dataset to
        n_samples (int): How many samples of the synthetic reasoning dataset to create (each sample contains 100 data points)
        specs (list[DatasetSpec]): specs for creating the synthetic reasoning dataset. Defaults to specs.
    
    Returns:
        None
    """
    
    n_failed=0
    for seed in random.sample(range(1_000_000), n_samples):
        data = reasoning_gym.create_dataset('composite', size=100, seed=seed, datasets=specs)

        with open(out_path, 'a') as f:
            for i, x in enumerate(data):
                assert data.score_answer(answer=x['answer'], entry=x) == 1.0
                assert type(x) == dict
                try:
                    x = {
                        'subject': x['metadata']['source_dataset'],
                        'text': [{'role': 'user', 'text': x['question']}, {'role': 'assistant', 'text': x['answer']}],
                        'data type': 'conv'
                    }
                    f.write(json.dumps(x) + '\n')
                except Exception as e:
                    n_failed += 1
                    print(f'[Fail {n_failed}# ]', e, x['metadata']['source_dataset'])



import random
from pydantic import BaseModel, Field
import os
import json
# from tqdm import tqdm
# from openai import OpenAI
import os
import copy
import re
from typing import Optional, Any
import vllm
from transformers import AutoTokenizer
from rich.progress import Progress, TextColumn, BarColumn, TimeElapsedColumn, ProgressColumn
from datetime import timedelta
from pathlib import Path

llm = None
llm_model_name = None

class Response(BaseModel):
    question: str = Field(description="What question is appropriate to this text?")
    answer: str = Field(description="Answer to the question")

sys_prompt = """
You are an assistant in a multi-turn Q&A about *The Witcher* universe.

Task:
1. Read the provided content.
2. Write ONE clear, unique question about it.
3. Write ONE concise, factual answer.

Rules:
- Use only the English language.
- Stay within Witcher lore (books, games, monsters, history, characters). 
- Never mention the Netflix series, actors, or production.
- Each Q must be different from all "Already asked questions by the user".
- No duplicates, paraphrases, or overlapping questions. If stuck, pivot to a new angle.
- Keep Q&A short, precise, and directly tied to the content.

Output:
Return only ONE valid JSON object, ending with `}`:

{
    "question": "<concise question>",
    "answer": "<concise factual answer>"
}

Content:
{content}

Already asked questions by the user:
"""

new_context_prompt = """
New content has been added. Based only on this NEW CONTENT:

1. Ask ONE clear, unique question.
2. Provide ONE concise, factual answer.

Rules:
- Must be about the NEW CONTENT.
- Never mention Netflix/actors/production.
- Do not repeat or rephrase any previous or "Already asked" questions.
- If no new unique question is possible, ask a short clarification instead.
- Keep Q&A minimal and factual.

Output:
Return only ONE valid JSON object, ending with `}`:

{
    "question": "<concise question>",
    "answer": "<concise factual answer>"
}

NEW CONTENT:
{random_content}
"""

random_user_prompt1 = \
"""Ask to elaborate."""

random_user_prompt2 = \
"""Ask to clarify."""

random_user_prompt3 = \
"""Ask to rephrase the answer."""

random_user_prompt4 = \
"""Cite some of the context and ask to summarize."""

random_user_prompt5 = \
"""Cite some of the context and ask to explain."""

random_user_prompt6 = \
"""Ask for a simpler answer."""

random_user_prompt7 = \
"""Ask if a a particular fact is true or false. The stated fact has to be based on the context and has to be true."""

random_user_prompt8 = \
"""Ask if a particular fact is true or false. The stated fact has to be based on the context and has to be false."""

random_user_prompt9 = \
"""Ask a very underspecific question."""

random_user_prompt10 = \
"""Ask for an example related to the answer."""

random_user_prompt11 = \
"""Ask why the answer is important."""

random_user_prompt12 = \
"""Ask how this relates to other events or characters."""

random_user_prompt13 = \
"""Ask for a comparison with another character, place, or event."""

random_user_prompt14 = \
"""Ask the assistant to list two more related facts or characters or events."""

random_user_prompt15 = \
"""Ask to put the answer into chronological context if possible."""

random_user_prompt16 = \
"""Ask to explain the political or historical significance of the answer."""


random_user_prompts = (random_user_prompt1, random_user_prompt2,
                       random_user_prompt3, random_user_prompt4,
                       random_user_prompt5, random_user_prompt6,
                       random_user_prompt7, random_user_prompt8,
                       random_user_prompt9, random_user_prompt10,
                       random_user_prompt11, random_user_prompt12,
                       random_user_prompt13, random_user_prompt14,
                       random_user_prompt15, random_user_prompt16)

class StepEtaColumn(ProgressColumn):
    """ETA updates only when task.completed changes"""
    def __init__(self):
        super().__init__()
        self.last_completed = {}
        self.cached_eta = {}

    def render(self, task):
        # Initialize per-task tracking
        if task.id not in self.last_completed:
            self.last_completed[task.id] = -1
            self.cached_eta[task.id] = "ETA: --:--"

        # Update ETA only if completed count changed
        if task.completed != self.last_completed[task.id] and task.completed > 0:
            percent = task.completed / task.total if task.total else 0
            elapsed = task.elapsed or 0
            if percent > 0:
                eta_seconds = elapsed / percent - elapsed
                self.cached_eta[task.id] = f"ETA: {str(timedelta(seconds=int(eta_seconds)))}"
            self.last_completed[task.id] = task.completed

        return TextColumn(self.cached_eta[task.id]).render(task)

def best_effort_parse(text: str, schema: Optional[BaseModel] = None) -> dict[str, Any]:
    text = text.strip()

    try:
        data = json.loads(text)
        if schema:
            return schema(**data).model_dump()
        return data
    except Exception:
        pass

    try:
        match = re.search(r"\{.*\}", text, re.DOTALL)
        if match:
            candidate = match.group(0)
            data = json.loads(candidate)
            if schema:
                return schema(**data).model_dump()
            return data
    except Exception:
        pass

    candidate = re.sub(r",\s*}", "}", text)
    candidate = re.sub(r",\s*]", "]", candidate)
    try:
        data = json.loads(candidate)
        if schema:
            return schema(**data).dict()
        return data
    except Exception:
        pass

    return {"question": "", "answer": ""}

def vllm_multi_turn(paths:list[Path],
                    save_to:Path='./convos',
                    batch_size:int=256,
                    system_prompt:str=sys_prompt,
                    n_turns:int=5,
                    template:BaseModel=Response,
                    model:str="unsloth/Meta-Llama-3.1-8B-Instruct",
                    prob_chance_new_context:float=0.3):
    """
    Generate multi-turn conversations
    
    Args:
        paths (list[Path]): paths to .txt files containing texts to create multi turn question answer pairs
        save_to (Path): folder to save generated multi turn question answer pairs to. Defaults to './convos'.
        batch_size (int): batch size for the model. Defaults to 256.
        system_prompt (str): system prompt for the model. Defaults to sys_prompt.
        n_turns (int): number of turns to generate. Defaults to 5.
        template (BaseModel): template for the model. Defaults to Response.
        model (str): model to use. Defaults to "unsloth/Meta-Llama-3.1-8B-Instruct".
        prob_chance_new_context (float): probability of starting a new context for the question answer pairs. Defaults to 0.3.
    
    Returns:
        None
    """

    global llm, llm_model_name


    if llm is None or llm_model_name != model:
        llm = vllm.LLM(model=model, max_model_len=4096, gpu_memory_utilization=0.8)
        llm_model_name = model

    sampling = vllm.SamplingParams(
                max_tokens=2048,
            )

    tokenizer = AutoTokenizer.from_pretrained(model)

    # client = OpenAI(base_url=f"http://0.0.0.0:{vllm_port}/v1", api_key="dummy")

    # bar = tqdm(total=len(paths) * n_turns, desc=f"{model} conversations")
    os.makedirs(save_to, exist_ok=True)

    progress = Progress(
                        TextColumn("{task.description}"),
                        BarColumn(),
                        TextColumn("{task.completed}/{task.total}"),
                        TextColumn("{task.percentage:>3.0f}%"),
                        TimeElapsedColumn(),
                        StepEtaColumn(),
                        )
    
    writer = progress.add_task(f"{model.split('/')[1]}", total=len(paths) * n_turns)

    qa = []
    model_convo_batched = []
    contexts = []

    with progress:
        for p in paths:

            with open(p, 'r') as f:
                path_content = f.read()

            model_convo = [{'role': 'system', 'content': system_prompt.replace('{content}', path_content)}]
            model_convo_batched.append(model_convo)
            contexts.append(os.path.basename(p).split('.')[0])

        for turn in range(n_turns):
        
            batched_output = []

            for i in range(0, len(model_convo_batched), batch_size):

                chat_template_batch  = [
                                        tokenizer.apply_chat_template(
                                            prompt,
                                            tokenize=False,
                                            add_generation_prompt=True
                                        ) for prompt in model_convo_batched[i:i+batch_size]
                                        ]

                out = llm.generate(chat_template_batch, sampling_params=sampling)
                out = [best_effort_parse(o.outputs[0].text, template) for o in out]
                batched_output.extend(out)

                # batch = model_convo_batched[i:i+batch_size]
                # try:
                #     responses = [
                #         client.beta.chat.completions.parse(
                #             model=model,
                #             messages=conv,
                #             response_format=template,
                #             timeout=timeout
                #         )
                #         for conv in batch
                #     ]
                #     parsed = [resp.choices[0].message.parsed.model_dump() for resp in responses]
                
                # except:
                #     bar.write('Failed to generate response - timeout')
                #     parsed = [{'question': '', 'answer': ''} for _ in range(len(batch))]

                # batched_output.extend(parsed)

                progress.update(writer, advance=len(out))
            
            for i, b in enumerate(batched_output):
                b['context'] = contexts[i]

            qa.append(batched_output)

            for i in range(len(model_convo_batched)):

                model_convo_batched[i].append({'role':'assistant', 'content':f'Question: {batched_output[i]["question"]}\nAnswer: {batched_output[i]["answer"]}\n\n'})

                if (random.randint(0, 100) < prob_chance_new_context * 100) and (turn > 0):

                        random_path = random.choice(paths)
                        with open(random_path, 'r') as f:
                            random_content = f.read()

                        model_convo_batched[i].append({'role':'user', 'content':new_context_prompt.replace("{random_content}", random_content)})
                        contexts[i] = os.path.basename(random_path).split('.')[0]

                else:
                    model_convo_batched[i].append({'role':'user', 'content':random.choice(random_user_prompts)})


    qa_gathered = [[] for _ in range(len(paths))]

    for turn in qa:
        assert len(turn) == len(paths)

        for i in range(len(paths)):
            if (not turn[i]['question'] == '') and (not turn[i]['answer'] == ''):
                qa_gathered[i].append(turn[i])
    
    for i, q in enumerate(qa_gathered):
        if len(q) > 0:
            
            path = f'{save_to}/{os.path.basename(paths[i]).split(".")[0]}'
            path_ = copy.deepcopy(path)
            i = 1
            while os.path.exists(path_+'.json'):
                path_ = copy.deepcopy(path) + f'_{i}'
                i += 1

            with open(f'{path_}.json', 'w') as f:
                json.dump(q, f, indent=2)



from .activations import get_activation, Dynamic_erf, DynamicTanh
from .SMoE import SMoE, SwiGLU
from .RoPE import RoPE
from .sliding_window_attention import (
                                    SlidingWindowAttention,
                                    create_dynamic_block_mask,
                                    create_static_block_mask,
                                    sliding_window_causal
                                      )
from .BERT_attention import BertAttention

__all__ = [
    'get_activation',
    'SMoE',
    'SwiGLU',
    'RoPE',
    'SlidingWindowAttention',
    'create_dynamic_block_mask',
    'create_static_block_mask',
    'sliding_window_causal',
    'BertAttention',
    'Dynamic_erf',
    'DynamicTanh'
]

try:
    from .SMoE_megablocks import MegablockMoE, MegablockdMoE

    __all__ += [
        "MegablockMoE",
        "MegablockdMoE",
    ]
except ImportError:
    pass


import os
import json
import re
from pathlib import Path

def multi_turn_gather(input_path:Path, save_to:Path):
    """
    create synthetic multi turn conversations about given topics with vllm
    
    Args:
        input_path (Path): path to the folder containing the .json files, they can be nested
        save_to (Path): path to save the synthetic multi turn dataset to
    
    Returns:
        None
    """
    files = []
    for main_path, subfolders, _files in os.walk(input_path):
        for _file in _files:
            if _file.endswith(".json"):
                files.append(os.path.join(main_path, _file))

    for file in files:
        with open(os.path.join(file), 'r') as f:
            qa_list = json.load(f)
            out_data = []
            contexts = []
            if not isinstance(qa_list, list):
                qa_list = [qa_list]
            for qa in qa_list:
                out_data.extend([{'role':'user', 'content': qa['question']}, {'role': 'assistant', 'content': qa['answer']}])
                contexts.append(qa['context'])
            out = {'subject': re.sub(r'_\d+$', '', file.split('/')[-1].split('.')[0]),
                'text': out_data,
                'data type': 'conv',
                'source': 'fandom',
                'metadata': {'contexts': contexts}}
                
            with open(save_to, 'a') as f:
                f.write(json.dumps(out) + '\n')



import torch.nn as nn
import torch
from torch.utils.data import DataLoader, TensorDataset, Dataset, IterableDataset
from ..Cirilla_model.dataloader import GenericDataset
import json
import polars as pl
import random
from types import MethodType

class Task(Dataset):
    def __init__(self, name, data, batch_size=512):
        super().__init__()
        self.name = name
        self.batch_size = batch_size

        data = pl.DataFrame({'x': data[0], 'y': data[1]})
        data = data.sample(fraction=1.0, shuffle=True)

        n_rows = data.shape[0]
        n_parts = n_rows // batch_size
        if n_rows % batch_size != 0:
            n_parts += 1

        self.data = tuple(
            data[part * batch_size : (part + 1) * batch_size] 
            for part in range(n_parts)
        )

        self.n_parts = n_parts

    def __len__(self):
        return self.n_parts

    def __getitem__(self, idx):
        data = self.data[idx]
        shape = data.shape[0] // 2

        support_smiles = data[shape:, 0].to_list()
        support_labels = torch.tensor(data[shape:, 1].to_list())

        query_smiles = data[:shape, 0].to_list()
        query_labels = torch.tensor(data[:shape, 1].to_list())

        return {
                'support_data': support_smiles,
                'support_labels': support_labels,
                'query_data': query_smiles,
                'query_labels': query_labels
                }

class MamlPretrainingDataset(IterableDataset, GenericDataset):
    def __init__(self, batch_size=512, **kwargs):
        super().__init__(**kwargs)
        self.batch_size = batch_size
        self._get_data()

    def _get_data(self):
        
        self.task_data = {}

        for i, p in enumerate(self.path):
            texts = []
            labels = []
            with open(p, 'r') as f:

                for line in f:
                    line = json.loads(line)
                    assert line['data type'] == 'bert', 'MAML supports only BERT data'

                    text = line['text']

                    if self.prefix_tokens is not None:
                        text = "".join(self.prefix_tokens) + text

                    if self.suffix_tokens is not None:
                        text += "".join(self.suffix_tokens)

                    texts.append(text)
                    labels.append(line['label'])
            
            self.task_data[p.split('/')[-1].strip('.jsonl') + str(i)] = (texts, labels)

    def shuffle_tasks(self):
        self.task_data = list(self.task_data.items())
        random.shuffle(self.task_data)
        self.task_data = dict(self.task_data)

    def __len__(self):
        return len(self.task_data)

    def __iter__(self):
        for k, v in self.task_data.items():
            yield Task(k, v, self.batch_size)
        

def _fine_tune(self,
                train_texts: list[str],
                train_labels: list[int],
                test_texts: list[str],
                test_labels: list[int],
                epochs: int = 50,
                batch_size: int = 16,
                learning_rate: float = 0.1,
                verbose: bool = False):

    tokenized_train = self.tokenizer(train_texts, return_tensors='pt', padding='max_length', truncation=True, max_length=self.max_len)
    tokenized_test = self.tokenizer(test_texts, return_tensors='pt', padding='max_length', truncation=True, max_length=self.max_len)

    train_ids = tokenized_train['input_ids']
    train_masks = tokenized_train['attention_mask']

    test_ids = tokenized_test['input_ids']
    test_masks = tokenized_test['attention_mask']

    train_labels_tensor = torch.tensor(train_labels, dtype=torch.float32)
    test_labels_tensor = torch.tensor(test_labels, dtype=torch.float32)

    train_dataset = TensorDataset(train_ids, train_masks, train_labels_tensor)
    train_loader = DataLoader(train_dataset, batch_size=batch_size, shuffle=True)

    test_dataset = TensorDataset(test_ids, test_masks, test_labels_tensor)
    test_loader = DataLoader(test_dataset, batch_size=batch_size, shuffle=True)

    optimizer = torch.optim.AdamW(self.model.parameters(), lr=learning_rate)

    for epoch in range(epochs):
        train_loss = 0
        valid_loss = 0
        nt = 0
        nv = 0
        for phase in ['train', 'valid']:

            if phase == 'train':
                self.model.train()

                for _ in range(self.num_inner_steps):

                    for batch_ids, masks, labels in train_loader:

                        predictions = self.model(batch_ids.to(self.device), masks.to(self.device))

                        inner_loss = self.loss_fn(predictions, labels.to(self.device, dtype=self.model.args.dtype if type(self.loss_fn) == nn.BCELoss else torch.int64))

                        optimizer.zero_grad()
                        inner_loss.backward()
                        optimizer.step()

                        b = batch_ids.shape[0]
                        nt += b
                        train_loss += inner_loss.item()*b
                
            else:
                self.model.eval()
                with torch.no_grad():
                    for batch_ids, masks, labels in test_loader:

                        predictions = self.model(batch_ids.to(self.device), masks.to(self.device))
                        loss = self.loss_fn(predictions, labels.to(self.device, dtype=self.model.args.dtype if type(self.loss_fn) == nn.BCELoss else torch.int64))

                        b = batch_ids.shape[0]
                        nv += b
                        valid_loss += loss.item()*b

        if verbose:
            print(f"Epoch {epoch}, train loss: {train_loss/nt:.4f}, valid loss: {valid_loss/nv:.4f}")

    return self

class MagMaxMAMLTrainer:
    def __init__(self,
                model,
                tokenizer,
                meta_lr=0.1,
                inner_lr=0.01,
                num_inner_steps=5,
                max_len=512):

        self.tokenizer = tokenizer
        self.model = model
        self.device = getattr(model.args, 'device')

        self.inner_lr = inner_lr
        self.meta_lr = meta_lr
        self.num_inner_steps = num_inner_steps

        self.loss_fn = nn.BCELoss()
        self.max_len = max_len

        self.fine_tune = MethodType(_fine_tune, self)
    
    def meta_train(self,
                    meta_train_tasks: list[dict[str, list]],
                    epochs: int = 50):

        for epoch in range(epochs):
            inner_loss_epoch = 0
            n=0

            model_zero_params = self.model.state_dict()
            weight_update = {k: torch.zeros_like(v) for k, v in model_zero_params.items()}

            adapted_model = type(self.model)(self.model.args).to(self.device, dtype=self.model.args.dtype)
            adapted_model.load_state_dict(self.model.state_dict())
            inner_optimizer = torch.optim.AdamW(adapted_model.parameters(), lr=self.inner_lr)

            meta_train_tasks.shuffle_tasks()

            for task in meta_train_tasks:

                for _ in range(self.num_inner_steps):
                
                    for n_iter, batch in enumerate(task):
                        support_data = batch['support_data']
                        support_labels = batch['support_labels']
                        query_data = batch['query_data']
                        query_labels = batch['query_labels']

                        if support_labels.shape[0] == 0 or query_labels.shape[0] == 0:
                            if (n_iter+1) != task.n_parts:
                                print(f"Skipping empty task: {task.name} at iter: {n_iter+1}/{task.n_parts}")

                            continue

                        supports = self.tokenizer(support_data, return_tensors='pt', padding='max_length', truncation=True, max_length=self.max_len)
                        support_ids, support_masks = supports['input_ids'], supports['attention_mask']

                        querys = self.tokenizer(query_data, return_tensors='pt', padding='max_length', truncation=True, max_length=self.max_len)
                        query_ids, query_masks = querys['input_ids'], querys['attention_mask']

                        ids = torch.cat((support_ids, query_ids), dim=0).to(self.device)
                        masks = torch.cat((support_masks, query_masks), dim=0).to(self.device)
                        labels = torch.cat((support_labels, query_labels), dim=0).to(self.device, dtype=self.model.args.dtype if type(self.loss_fn) == nn.BCELoss else torch.int64)

                        predictions = adapted_model(ids, masks)

                        inner_loss = self.loss_fn(predictions, labels)

                        inner_optimizer.zero_grad(set_to_none=True)
                        inner_loss.backward()
                        inner_optimizer.step()

                        inner_loss_epoch += inner_loss.item()
                        n+=1

                with torch.no_grad():
                    adapted_params = adapted_model.state_dict()
                    param_deltas = {key: value - model_zero_params[key] for key, value in adapted_params.items()}

                    for key in weight_update.keys():
                        current_update = weight_update[key]
                        new_update = param_deltas[key]

                        # get biggest absolute changes mask
                        mask = torch.abs(new_update) > torch.abs(current_update)
                        current_update[mask] = new_update[mask]

                        weight_update[key] = current_update

            with torch.no_grad():
                
                for key in model_zero_params.keys():
                    model_zero_params[key] += self.meta_lr * weight_update[key]

                self.model.load_state_dict(model_zero_params)

            print(f"Epoch {epoch+1}, Tasks-specific Loss: {inner_loss_epoch/n:.4f}")

    def fine_tune_native(self,
                    train_texts: list[str],
                    train_labels: list[int],
                    test_texts: list[str],
                    test_labels: list[int],
                    epochs: int = 50,
                    batch_size: int = 16,
                    learning_rate: float = 0.1,
                    verbose: bool = False):

        tokenized_train = self.tokenizer(train_texts, return_tensors='pt', padding='max_length', truncation=True, max_length=self.max_len)
        tokenized_test = self.tokenizer(test_texts, return_tensors='pt', padding='max_length', truncation=True, max_length=self.max_len)

        train_ids = tokenized_train['input_ids']
        train_masks = tokenized_train['attention_mask']

        test_ids = tokenized_test['input_ids']
        test_masks = tokenized_test['attention_mask']

        train_labels_tensor = torch.tensor(train_labels, dtype=torch.float32)
        test_labels_tensor = torch.tensor(test_labels, dtype=torch.float32)

        train_dataset = TensorDataset(train_ids, train_masks, train_labels_tensor)
        train_loader = DataLoader(train_dataset, batch_size=batch_size, shuffle=True)

        test_dataset = TensorDataset(test_ids, test_masks, test_labels_tensor)
        test_loader = DataLoader(test_dataset, batch_size=batch_size, shuffle=True)

        for epoch in range(epochs):
            train_loss = 0
            valid_loss = 0
            nt = 0
            nv = 0
            for phase in ['train', 'valid']:

                if phase == 'train':
                    self.model.train()

                    adapted_model = type(self.model)(self.model.args).to(self.device)
                    adapted_model.load_state_dict(self.model.state_dict())
                    inner_optimizer = torch.optim.SGD(adapted_model.parameters(), lr=self.inner_lr)

                    for _ in range(self.num_inner_steps):

                        for batch_ids, masks, labels in train_loader:

                            predictions = adapted_model(batch_ids.to(self.device), masks.to(self.device))

                            inner_loss = self.loss_fn(predictions, labels.to(self.device, dtype=self.model.args.dtype if type(self.loss_fn) == nn.BCELoss else torch.int64))

                            inner_optimizer.zero_grad()
                            inner_loss.backward()
                            inner_optimizer.step()

                            b = batch_ids.shape[0]
                            nt += b
                            train_loss += inner_loss.item()*b

                    with torch.no_grad(): # for one task this is basically reduced to reptile
                        original_param_dict = self.model.state_dict()
                        adapted_param_dict = adapted_model.state_dict()

                        for key in original_param_dict.keys():
                            original_param_dict[key] = original_param_dict[key] + learning_rate * (adapted_param_dict[key] - original_param_dict[key])

                        self.model.load_state_dict(original_param_dict)
                    
                else:
                    self.model.eval()
                    with torch.no_grad():
                        for batch_ids, masks, labels in test_loader:

                            predictions = self.model(batch_ids.to(self.device), masks.to(self.device))
                            loss = self.loss_fn(predictions, labels.to(self.device, dtype=self.model.args.dtype if type(self.loss_fn) == nn.BCELoss else torch.int64))

                            b = batch_ids.shape[0]
                            nv += b
                            valid_loss += loss.item()*b

            if verbose:
                print(f"Epoch {epoch}, train loss: {train_loss/nt:.4f}, valid loss: {valid_loss/nv:.4f}")

        return self

class ReptileTrainer:
    def __init__(self,
                model,
                tokenizer,
                meta_lr=0.1,
                inner_lr=0.01,
                num_inner_steps=5,
                max_len=512):

        self.tokenizer = tokenizer
        self.model = model

        self.device = getattr(model.args, 'device')

        self.inner_lr = inner_lr
        self.meta_lr = meta_lr
        self.num_inner_steps = num_inner_steps

        self.loss_fn = nn.BCELoss()
        self.max_len = max_len

        self.fine_tune = MethodType(_fine_tune, self)
    
    def meta_train(self,
                    meta_train_tasks: list[dict[str, list]],
                    epochs: int = 50):

        for epoch in range(epochs):
            meta_train_loss = 0
            n=0

            meta_train_tasks.shuffle_tasks()

            for task in meta_train_tasks:

                adapted_model = type(self.model)(self.model.args).to(self.device)
                adapted_model.load_state_dict(self.model.state_dict())
                inner_optimizer = torch.optim.SGD(adapted_model.parameters(), lr=self.inner_lr)

                for _ in range(self.num_inner_steps):

                    for n_iter, batch in enumerate(task):

                        support_data = batch['support_data']
                        support_labels = batch['support_labels']
                        query_data = batch['query_data']
                        query_labels = batch['query_labels']

                        if support_labels.shape[0] == 0 or query_labels.shape[0] == 0:
                            if (n_iter+1) != task.n_parts:
                                print(f"Skipping empty task: {task.name} at iter: {n_iter+1}/{task.n_parts}")

                            continue

                        supports = self.tokenizer(support_data, return_tensors='pt', padding='max_length', truncation=True, max_length=self.max_len)
                        support_ids, support_masks = supports['input_ids'], supports['attention_mask']

                        querys = self.tokenizer(query_data, return_tensors='pt', padding='max_length', truncation=True, max_length=self.max_len)
                        query_ids, query_masks = querys['input_ids'], querys['attention_mask']

                        ids = torch.cat((support_ids, query_ids), dim=0).to(self.device)
                        masks = torch.cat((support_masks, query_masks), dim=0).to(self.device)
                        labels = torch.cat((support_labels, query_labels), dim=0).to(self.device, dtype=self.model.args.dtype if type(self.loss_fn) == nn.BCELoss else torch.int64)

                        predictions = adapted_model(ids, masks)

                        inner_loss = self.loss_fn(predictions, labels)

                        inner_optimizer.zero_grad(set_to_none=True)
                        inner_loss.backward()
                        inner_optimizer.step()

                        meta_train_loss += inner_loss.item()
                        n+=1

                with torch.no_grad():
                    original_param_dict = self.model.state_dict()
                    adapted_param_dict = adapted_model.state_dict()

                    for key in original_param_dict.keys():
                        original_param_dict[key] = original_param_dict[key] + self.meta_lr * (adapted_param_dict[key] - original_param_dict[key])

                    self.model.load_state_dict(original_param_dict)

            print(f"Epoch {epoch+1}, Tasks-specific Loss: {meta_train_loss/n:.4f}")

    def fine_tune_native(self,
                    train_texts: list[str],
                    train_labels: list[int],
                    test_texts: list[str],
                    test_labels: list[int],
                    epochs: int = 50,
                    batch_size: int = 16,
                    learning_rate: float = 0.1,
                    verbose: bool = False):

        tokenized_train = self.tokenizer(train_texts, return_tensors='pt', padding='max_length', truncation=True, max_length=self.max_len)
        tokenized_test = self.tokenizer(test_texts, return_tensors='pt', padding='max_length', truncation=True, max_length=self.max_len)

        train_ids = tokenized_train['input_ids']
        train_masks = tokenized_train['attention_mask']

        test_ids = tokenized_test['input_ids']
        test_masks = tokenized_test['attention_mask']

        train_labels_tensor = torch.tensor(train_labels, dtype=torch.float32)
        test_labels_tensor = torch.tensor(test_labels, dtype=torch.float32)

        train_dataset = TensorDataset(train_ids, train_masks, train_labels_tensor)
        train_loader = DataLoader(train_dataset, batch_size=batch_size, shuffle=True)

        test_dataset = TensorDataset(test_ids, test_masks, test_labels_tensor)
        test_loader = DataLoader(test_dataset, batch_size=batch_size, shuffle=True)

        for epoch in range(epochs):
            train_loss = 0
            valid_loss = 0
            nt = 0
            nv = 0
            for phase in ['train', 'valid']:
                
                if phase == 'train':
                    self.model.train()

                    adapted_model = type(self.model)(self.model.args).to(self.device)
                    adapted_model.load_state_dict(self.model.state_dict())
                    inner_optimizer = torch.optim.SGD(adapted_model.parameters(), lr=self.inner_lr)

                    for _ in range(self.num_inner_steps):

                        for batch_ids, masks, labels in train_loader:

                            predictions = adapted_model(batch_ids.to(self.device), masks.to(self.device))

                            inner_loss = self.loss_fn(predictions, labels.to(self.device, dtype=self.model.args.dtype if type(self.loss_fn) == nn.BCELoss else torch.int64))

                            inner_optimizer.zero_grad()
                            inner_loss.backward()
                            inner_optimizer.step()

                            b = batch_ids.shape[0]
                            nt += b
                            train_loss += inner_loss.item()*b

                    with torch.no_grad():
                        original_param_dict = self.model.state_dict()
                        adapted_param_dict = adapted_model.state_dict()

                        for key in original_param_dict.keys():
                            original_param_dict[key] = original_param_dict[key] + learning_rate * (adapted_param_dict[key] - original_param_dict[key])

                        self.model.load_state_dict(original_param_dict)
                    
                else:
                    self.model.eval()
                    with torch.no_grad():
                        for batch_ids, masks, labels in test_loader:

                            predictions = self.model(batch_ids.to(self.device), masks.to(self.device))
                            loss = self.loss_fn(predictions, labels.to(self.device, dtype=self.model.args.dtype if type(self.loss_fn) == nn.BCELoss else torch.int64))

                            b = batch_ids.shape[0]
                            nv += b
                            valid_loss += loss.item()*b

            if verbose:
                print(f"Epoch {epoch}, train loss: {train_loss/nt:.4f}, valid loss: {valid_loss/nv:.4f}")

        return self

class MAMLBinaryAdapterTrainer:
    def __init__(self,
                model,
                tokenizer,
                meta_lr=0.001,
                inner_lr=0.01,
                num_inner_steps=5,
                max_len=512):
        
        self.tokenizer = tokenizer
        self.model = model

        self.device = getattr(model.args, 'device')

        self.embedding_dim = self.model.args.dim

        self.classifier = nn.Sequential(
            nn.Linear(self.embedding_dim, 64),
            nn.SiLU(),
            nn.Linear(64, 32),
            nn.SiLU(),
            nn.Linear(32, 1),
            nn.Sigmoid()
        ).to(self.device)

        self.meta_lr = meta_lr
        self.inner_lr = inner_lr
        self.num_inner_steps = num_inner_steps

        self.meta_optimizer = torch.optim.AdamW([
            *self.model.parameters(),
            *self.classifier.parameters()
        ], lr=self.meta_lr)

        self.loss_fn = nn.BCELoss()
        self.max_len = max_len

    def _get_embeddings(self, texts: list[str]) -> torch.Tensor:

        inputs = self.tokenizer(texts, return_tensors='pt', padding='max_length', truncation=True, max_length=self.max_len)

        with torch.no_grad():
            embeddings = self.model(inputs['input_ids'].to(self.device), inputs['attention_mask'].to(self.device))

        return embeddings

    def _inner_loop_adaptation(self,
                                support_embeddings: torch.Tensor,
                                support_labels: torch.Tensor) -> tuple[nn.Module, float]:

        adapted_classifier = type(self.classifier)(*self.classifier._modules.values()).to(self.device, dtype=self.model.args.dtype)
        adapted_classifier.load_state_dict(self.classifier.state_dict())

        inner_optimizer = torch.optim.SGD(adapted_classifier.parameters(), lr=self.inner_lr)

        for _ in range(self.num_inner_steps):
            predictions = adapted_classifier(support_embeddings)
            inner_loss = self.loss_fn(predictions.view(-1), support_labels)

            inner_optimizer.zero_grad(set_to_none=True)
            inner_loss.backward(retain_graph=True)
            inner_optimizer.step()

        return adapted_classifier, inner_loss.item()
    
    def meta_train(self,
                    meta_train_tasks: list[dict[str, list]],
                    epochs: int = 50):

        for epoch in range(epochs):
            meta_train_loss = 0
            n=0

            meta_train_tasks.shuffle_tasks()
            
            for task in meta_train_tasks:

                for n_iter, batch in enumerate(task):

                    support_data = batch['support_data']
                    support_labels = batch['support_labels'].to(self.device, dtype=self.model.args.dtype)
                    query_data = batch['query_data']
                    query_labels = batch['query_labels'].to(self.device, dtype=self.model.args.dtype)

                    if support_labels.shape[0] == 0 or query_labels.shape[0] == 0:
                        if (n_iter+1) != task.n_parts:
                            print(f"Skipping empty task: {task.name} at iter: {n_iter+1}/{task.n_parts}")

                        continue

                    support_embeddings = self.tokenizer(support_data, return_tensors='pt', padding='max_length', truncation=True, max_length=self.max_len)
                    support_embeddings = self.model(support_embeddings['input_ids'].to(self.device), support_embeddings['attention_mask'].to(self.device))

                    query_embeddings = self.tokenizer(query_data, return_tensors='pt', padding='max_length', truncation=True, max_length=self.max_len)
                    query_embeddings = self.model(query_embeddings['input_ids'].to(self.device), query_embeddings['attention_mask'].to(self.device))

                    adapted_classifier, _ = self._inner_loop_adaptation(support_embeddings, support_labels)

                    query_predictions = adapted_classifier(query_embeddings)
                    meta_loss = self.loss_fn(query_predictions.view(-1), query_labels)

                    self.meta_optimizer.zero_grad(set_to_none=True)
                    meta_loss.backward()
                    self.meta_optimizer.step()

                    meta_train_loss += meta_loss.item()
                    n+=1

                print(f"Epoch {epoch+1}, Tasks-specific Loss: {meta_train_loss/n:.4f}")

    def fine_tune(self,
                    train_texts: list[str],
                    train_labels: list[int],
                    test_texts: list[str],
                    test_labels: list[int],
                    epochs: int = 50,
                    batch_size: int = 16,
                    learning_rate: float = 1e-2,
                    verbose: bool = False):

        train_embeddings = self._get_embeddings(train_texts).detach()
        test_embeddings = self._get_embeddings(test_texts).detach()

        train_labels_tensor = torch.tensor(train_labels, dtype=self.model.args.dtype)
        test_labels_tensor = torch.tensor(test_labels, dtype=self.model.args.dtype)

        train_dataset = TensorDataset(train_embeddings, train_labels_tensor)
        train_loader = DataLoader(train_dataset, batch_size=batch_size, shuffle=True)

        test_dataset = TensorDataset(test_embeddings, test_labels_tensor)
        test_loader = DataLoader(test_dataset, batch_size=batch_size, shuffle=True)

        optimizer = torch.optim.AdamW(self.classifier.parameters(), lr=learning_rate)

        for epoch in range(epochs):
            train_loss = 0
            valid_loss = 0
            nt = 0
            nv = 0
            for phase in ['train', 'valid']:
                
                if phase == 'train':

                    self.classifier.train()
                    for batch_embeddings, batch_labels in train_loader:
                        batch_embeddings = batch_embeddings.to(self.device)
                        batch_labels = batch_labels.to(self.device)

                        predictions = self.classifier(batch_embeddings).view(-1)
                        loss = self.loss_fn(predictions, batch_labels)

                        optimizer.zero_grad()
                        loss.backward()
                        optimizer.step()

                        b = batch_embeddings.shape[0]
                        nt += b
                        train_loss += loss.item()*b
                    
                else:
                    self.classifier.eval()
                    with torch.no_grad():
                        for batch_embeddings, batch_labels in test_loader:
                            batch_embeddings = batch_embeddings.to(self.device)
                            batch_labels = batch_labels.to(self.device)

                            predictions = self.classifier(batch_embeddings).view(-1)
                            loss = self.loss_fn(predictions, batch_labels)

                            b = batch_embeddings.shape[0]
                            nv += b
                            valid_loss += loss.item()*b

            if verbose:
                print(f"Epoch {epoch}, train loss: {train_loss/nt:.4f}, valid loss: {valid_loss/nv:.4f}")

        return self


from kernels import get_kernel
from functools import lru_cache
import torch
import torch.nn as nn

@lru_cache(maxsize=8)
def get_activation(path: str = "kernels-community/activation"):
    return get_kernel(path)

class Dynamic_erf(nn.Module): # Stronger Normalization-Free Transformers Mingzhi Chen, Taiming Lu, Jiachen Zhu, Mingjie Sun, and Zhuang Liu. https://github.com/zlab-princeton/Derf/blob/main/language%20model/model_derf.patch
    def __init__(self, normalized_shape, alpha_init_value=0.5, shift_init_value=0.0):
        super().__init__()
        self.normalized_shape = normalized_shape
        self.alpha_init_value = alpha_init_value
        self.shift_init_value = shift_init_value

        self.alpha = nn.Parameter(torch.ones(1) * alpha_init_value)
        self.shift = nn.Parameter(torch.ones(1) * shift_init_value)
        self.weight = nn.Parameter(torch.ones(normalized_shape))
        self.bias = nn.Parameter(torch.zeros(normalized_shape))
        
    def forward(self, x):
        return self.weight * torch.erf(self.alpha * x + self.shift) + self.bias

class DynamicTanh(nn.Module): # Stronger Normalization-Free Transformers Mingzhi Chen, Taiming Lu, Jiachen Zhu, Mingjie Sun, and Zhuang Liu. https://github.com/zlab-princeton/Derf/blob/main/language%20model/model_derf.patch
    def __init__(self, normalized_shape, alpha_init_value=0.5):
        super().__init__()
        self.normalized_shape = normalized_shape
        self.alpha_init_value = alpha_init_value

        self.alpha = nn.Parameter(torch.ones(1) * alpha_init_value)
        self.weight = nn.Parameter(torch.ones(normalized_shape))
        self.bias = nn.Parameter(torch.zeros(normalized_shape))
        
    def forward(self, x):
        return self.weight * torch.tanh(self.alpha * x) + self.bias
